import pytest

import time

from moto import mock_xray_client, XRaySegment


def test_xray_tracing(caplog):
    @mock_xray_client
    def run_test():
        from terracotta import update_settings
        import terracotta.profile

        update_settings(XRAY_PROFILE=True)

        @terracotta.profile.trace("dummy")
        def func_to_trace():
            time.sleep(0.1)

        with XRaySegment():
            func_to_trace()

        with XRaySegment():
            with terracotta.profile.trace("dummy2"):
                time.sleep(0.1)

        for record in caplog.records:
            assert record.levelname != "ERROR"

        # sanity check, recording without starting a segment should fail
        func_to_trace()
        assert any(
            "cannot find the current segment" in rec.message for rec in caplog.records
        )

    run_test()


def test_xray_exception(caplog):
    @mock_xray_client
    def run_test():
        from terracotta import update_settings
        import terracotta.profile

        update_settings(XRAY_PROFILE=True)

        with XRaySegment():
            with pytest.raises(NotImplementedError):
                with terracotta.profile.trace("dummy") as subsegment:
                    raise NotImplementedError("foo")

        assert len(subsegment.cause["exceptions"]) == 1
        assert subsegment.cause["exceptions"][0].message == "foo"

    run_test()



"""drivers/base_classes.py

Base class for drivers.
"""

import contextlib
import functools
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from terracotta import exceptions

KeysType = Mapping[str, str]
MultiValueKeysType = Mapping[str, Union[str, List[str]]]
Number = TypeVar("Number", int, float)
T = TypeVar("T")


def requires_writable(fun: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(fun)
    def inner(self: MetaStore, *args: Any, **kwargs: Any) -> T:
        if self._WRITABLE:
            return fun(self, *args, **kwargs)
        else:
            raise exceptions.DatabaseNotWritableError("Database not writable")

    return inner


class MetaStore(ABC):
    """Abstract base class for all Terracotta metadata backends.

    Defines a common interface for all metadata backends.
    """

    _RESERVED_KEYS = ("limit", "page")
    _WRITABLE: bool = True

    @property
    @abstractmethod
    def db_version(self) -> str:
        """Terracotta version used to create the database."""
        pass

    @property
    @abstractmethod
    def key_names(self) -> Tuple[str, ...]:
        """Names of all keys defined by the database."""
        pass

    @abstractmethod
    def __init__(self, url_or_path: str) -> None:
        self.path = url_or_path

    @classmethod
    def _normalize_path(cls, path: str) -> str:
        """Convert given path to normalized version (that can be used for caching)"""
        return path

    @abstractmethod
    def create(
        self,
        keys: Sequence[str],
        *,
        key_descriptions: Optional[Mapping[str, str]] = None,
    ) -> None:
        """Create a new, empty database"""
        pass

    @abstractmethod
    def connect(self, verify: bool = True) -> contextlib.AbstractContextManager:
        """Context manager to connect to a given database and clean up on exit.

        This allows you to pool interactions with the database to prevent possibly
        expensive reconnects, or to roll back several interactions if one of them fails.
        """
        pass

    @abstractmethod
    def get_keys(self) -> OrderedDict:
        """Get all known keys and their fulltext descriptions."""
        pass

    @abstractmethod
    def get_datasets(
        self,
        where: Optional[MultiValueKeysType] = None,
        page: int = 0,
        limit: Optional[int] = None,
    ) -> Dict[Tuple[str, ...], Any]:
        """Get all known dataset key combinations matching the given constraints,
        and a path to retrieve the data
        """
        pass

    @abstractmethod
    def get_metadata(self, keys: KeysType) -> Optional[Dict[str, Any]]:
        """Return all stored metadata for given keys."""
        pass

    @abstractmethod
    def insert(
        self, keys: KeysType, path: str, *, metadata: Optional[Mapping[str, Any]] = None
    ) -> None:
        """Register a new dataset. This also populates the metadata database,
        if metadata is specified and not `None`."""
        pass

    @abstractmethod
    def delete(self, keys: KeysType) -> None:
        """Remove a dataset, including information from the metadata database."""
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.path}')"


class RasterStore(ABC):
    """Abstract base class for all Terracotta raster backends.

    Defines a common interface for all raster backends."""

    @abstractmethod
    # TODO: add accurate signature if mypy ever supports conditional return types
    def get_raster_tile(
        self,
        path: str,
        *,
        tile_bounds: Optional[Sequence[float]] = None,
        tile_size: Sequence[int] = (256, 256),
        preserve_values: bool = False,
        asynchronous: bool = False,
    ) -> Any:
        """Load a raster tile with given path and bounds."""
        pass

    @abstractmethod
    def compute_metadata(
        self,
        path: str,
        *,
        extra_metadata: Optional[Any] = None,
        use_chunks: Optional[bool] = None,
        max_shape: Optional[Sequence[int]] = None,
    ) -> Dict[str, Any]:
        """Compute metadata for a given input file"""
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"



"""drivers/mysql_meta_store.py

MySQL-backed metadata driver. Metadata is stored in a MySQL database.
"""

import functools
from typing import Optional, Mapping, Sequence

import sqlalchemy as sqla
from sqlalchemy.dialects.mysql import TEXT, VARCHAR
from terracotta.drivers.relational_meta_store import RelationalMetaStore


class MySQLMetaStore(RelationalMetaStore):
    """A MySQL-backed metadata driver.

    Stores metadata and paths to raster files in MySQL.

    Requires a running MySQL server.

    The MySQL database consists of 4 different tables:

    - ``terracotta``: Metadata about the database itself.
    - ``key_names``: Contains two columns holding all available keys and their description.
    - ``datasets``: Maps key values to physical raster path.
    - ``metadata``: Contains actual metadata as separate columns. Indexed via key values.

    This driver caches key names.
    """

    SQL_DIALECT = "mysql"
    SQL_DRIVER = "pymysql"
    SQL_TIMEOUT_KEY = "connect_timeout"

    _CHARSET = "utf8mb4"
    SQLA_STRING = functools.partial(VARCHAR, charset=_CHARSET)

    MAX_PRIMARY_KEY_SIZE = 767 // 4  # Max key length for MySQL is at least 767B
    DEFAULT_PORT = 3306

    def __init__(self, mysql_path: str) -> None:
        """Initialize the MySQLDriver.

        This should not be called directly, use :func:`~terracotta.get_driver` instead.

        Arguments:

            mysql_path: URL to running MySQL server, in the form
                ``mysql://username:password@hostname/database``

        """
        super().__init__(f"{mysql_path}?charset={self._CHARSET}")

        self.SQLA_METADATA_TYPE_LOOKUP["text"] = functools.partial(
            TEXT, charset=self._CHARSET
        )

        # raise an exception if database name is invalid
        if not self.url.database:
            raise ValueError("database must be specified in MySQL path")
        if "/" in self.url.database.strip("/"):
            raise ValueError("invalid database path")

    @classmethod
    def _normalize_path(cls, path: str) -> str:
        url = cls._parse_path(path)

        path = f"{url.drivername}://{url.host}:{url.port or cls.DEFAULT_PORT}/{url.database}"
        path = path.rstrip("/")
        return path

    def _create_database(self) -> None:
        engine = sqla.create_engine(
            self.url.set(
                database=""
            ),  # `.set()` returns a copy with changed parameters
            echo=False,
            future=True,
        )
        with engine.connect() as connection:
            connection.execute(sqla.text(f"CREATE DATABASE {self.url.database}"))
            connection.commit()

    def _initialize_database(
        self, keys: Sequence[str], key_descriptions: Optional[Mapping[str, str]] = None
    ) -> None:
        # total primary key length has an upper limit in MySQL
        self.SQL_KEY_SIZE = self.MAX_PRIMARY_KEY_SIZE // len(keys)
        super()._initialize_database(keys, key_descriptions)



"""migrations/__init__.py

Define available migrations.
"""

import os
import glob
import importlib


MIGRATIONS = {}

glob_pattern = os.path.join(os.path.dirname(__file__), "v*_*.py")

for modpath in glob.glob(glob_pattern):
    modname = os.path.basename(modpath)[: -len(".py")]
    mod = importlib.import_module(f"{__name__}.{modname}")
    assert all(
        hasattr(mod, attr) for attr in ("up_version", "down_version", "upgrade_sql")
    )
    assert mod.down_version not in MIGRATIONS
    MIGRATIONS[mod.down_version] = mod



from PIL import Image
import numpy as np

import pytest

from terracotta import exceptions


@pytest.mark.parametrize("resampling_method", ["nearest", "linear", "cubic", "average"])
def test_singleband_handler(use_testdb, raster_file_xyz, resampling_method):
    import terracotta

    terracotta.update_settings(RESAMPLING_METHOD=resampling_method)

    from terracotta.handlers import datasets, singleband

    settings = terracotta.get_settings()
    ds = datasets.datasets()

    for keys in ds:
        raw_img = singleband.singleband(keys, raster_file_xyz)
        img_data = np.asarray(Image.open(raw_img))
        assert img_data.shape == settings.DEFAULT_TILE_SIZE


def test_singleband_tile_size(use_testdb, raster_file_xyz):
    from terracotta.handlers import datasets, singleband

    ds = datasets.datasets()

    test_tile_size = (16, 32)

    for keys in ds:
        raw_img = singleband.singleband(keys, raster_file_xyz, tile_size=test_tile_size)
        img_data = np.asarray(Image.open(raw_img))
        assert img_data.shape == test_tile_size


def test_singleband_out_of_bounds(use_testdb):
    import terracotta
    from terracotta.handlers import datasets, singleband

    ds = datasets.datasets()

    for keys in ds:
        with pytest.raises(terracotta.exceptions.TileOutOfBoundsError):
            singleband.singleband(keys, (10, 0, 0))


def test_singleband_explicit_colormap(use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import singleband

    ds_keys = ["val21", "x", "val22"]
    nodata = 10000

    settings = terracotta.get_settings()
    driver = terracotta.get_driver(testdb)
    with driver.connect():
        tile_data = get_tile_data(
            driver,
            ds_keys,
            tile_xyz=raster_file_xyz,
            preserve_values=True,
            tile_size=settings.DEFAULT_TILE_SIZE,
        )

    # Get some values from the raster to use for colormap
    classes = np.unique(tile_data)
    classes = classes[:254]

    colormap = {}
    for i in range(classes.shape[0]):
        val = classes[i]
        color = val % 256
        colormap[val] = (color, color, color, color)
    colormap[nodata] = (100, 100, 100, 100)

    raw_img = singleband.singleband(ds_keys, raster_file_xyz, colormap=colormap)
    img_data = np.asarray(Image.open(raw_img).convert("RGBA"))

    # get unstretched data to compare to
    with driver.connect():
        tile_data = get_tile_data(
            driver,
            ds_keys,
            tile_xyz=raster_file_xyz,
            preserve_values=True,
            tile_size=img_data.shape[:2],
        )

    # check that labels are mapped to colors correctly
    for cmap_label, cmap_color in colormap.items():
        if cmap_label == nodata:
            # make sure nodata is still transparent
            assert np.all(img_data[tile_data == cmap_label, -1] == 0)
        else:
            assert np.all(img_data[tile_data == cmap_label] == np.asarray(cmap_color))

    # check that all data outside of labels is transparent
    keys_arr = np.array(list(colormap.keys()), dtype=np.int16)
    assert np.all(img_data[~np.isin(tile_data, keys_arr), -1] == 0)


def test_singleband_noxyz(use_testdb):
    from terracotta import get_settings
    from terracotta.handlers import singleband

    settings = get_settings()
    ds_keys = ["val21", "x", "val22"]

    raw_img = singleband.singleband(ds_keys)
    img_data = np.asarray(Image.open(raw_img))

    assert img_data.shape == settings.DEFAULT_TILE_SIZE


def test_singleband_stretch(use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import singleband

    ds_keys = ["val21", "x", "val22"]
    stretch_range = [0, 10000]

    raw_img = singleband.singleband(
        ds_keys,
        tile_xyz=raster_file_xyz,
        stretch_range=stretch_range,
    )
    img_data = np.asarray(Image.open(raw_img))

    # get unstretched data to compare to
    driver = terracotta.get_driver(testdb)

    with driver.connect():
        tile_data = get_tile_data(
            driver, ds_keys, tile_xyz=raster_file_xyz, tile_size=img_data.shape
        )

    # filter transparent values
    valid_mask = ~tile_data.mask
    assert np.all(img_data[~valid_mask] == 0)

    valid_img = img_data[valid_mask]
    valid_data = tile_data.compressed()

    assert np.all(valid_img[valid_data < stretch_range[0]] == 1)
    stretch_range_mask = (valid_data > stretch_range[0]) & (
        valid_data < stretch_range[1]
    )
    assert np.all(valid_img[stretch_range_mask] >= 1)
    assert np.all(valid_img[stretch_range_mask] <= 255)
    assert np.all(valid_img[valid_data > stretch_range[1]] == 255)


def test_singleband_stretch_percentile(use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import singleband

    ds_keys = ["val21", "x", "val22"]
    pct_stretch_range = ["p2", "p98"]

    raw_img = singleband.singleband(
        ds_keys,
        tile_xyz=raster_file_xyz,
        stretch_range=pct_stretch_range,
    )
    img_data = np.asarray(Image.open(raw_img))

    # get unstretched data to compare to
    driver = terracotta.get_driver(testdb)

    with driver.connect():
        tile_data = get_tile_data(
            driver, ds_keys, tile_xyz=raster_file_xyz, tile_size=img_data.shape
        )
        band_metadata = driver.get_metadata(ds_keys)

        stretch_range = [
            band_metadata["percentiles"][1],
            band_metadata["percentiles"][97],
        ]

    # filter transparent values
    valid_mask = ~tile_data.mask
    assert np.all(img_data[~valid_mask] == 0)

    valid_img = img_data[valid_mask]
    valid_data = tile_data.compressed()

    assert np.all(valid_img[valid_data < stretch_range[0]] == 1)
    stretch_range_mask = (valid_data > stretch_range[0]) & (
        valid_data < stretch_range[1]
    )
    assert np.all(valid_img[stretch_range_mask] >= 1)
    assert np.all(valid_img[stretch_range_mask] <= 255)
    assert np.all(valid_img[valid_data > stretch_range[1]] == 255)


@pytest.mark.parametrize(
    "stretch_range_params",
    [
        ["s2", "p98", "Invalid scale value"],
        ["pp2", "p98", "Invalid percentile value"],
        ["p", "p98", "Invalid percentile value"],
        ["2", "p8", "Invalid scale value"],
        [{}, "p98", "Invalid scale value"],
        ["p-2", "p98", "Invalid percentile, out of range"],
        ["p2", "p298", "Invalid percentile, out of range"],
    ],
)
def test_rgb_invalid_percentiles(use_testdb, stretch_range_params):
    from terracotta.handlers import singleband

    ds_keys = ["val21", "x", "val22"]

    stretch_range = stretch_range_params[:2]

    with pytest.raises(exceptions.InvalidArgumentsError) as err:
        singleband.singleband(
            ds_keys,
            stretch_range=stretch_range,
        )
    assert stretch_range_params[2] in str(err.value)



import os

REMOTE_DRIVERS = ("sqlite-remote", "mysql")


def check_integrity(zappa_cli):
    command = zappa_cli.command
    if command not in ("deploy", "update"):
        return

    env = zappa_cli.aws_environment_variables or {}

    db_provider = env.get("TC_DRIVER_PROVIDER")
    if db_provider not in REMOTE_DRIVERS:
        raise ValueError(
            f"TC_DRIVER_PROVIDER environment variable must be one of {REMOTE_DRIVERS}"
        )

    db_path = env.get("TC_DRIVER_PATH")
    if not db_path:
        raise ValueError("TC_DRIVER_PATH environment variable must be set")

    try:
        from terracotta import get_driver, exceptions
    except ImportError as exc:
        raise RuntimeError(
            "Terracotta must be installed before deployment (e.g. via `pip install .`)"
        ) from exc

    os.environ.update(env)
    driver = get_driver(db_path, provider=db_provider)

    # this checks if DB is reachable, readable, and whether its version matches
    try:
        with driver.connect():
            some_dataset = next(iter(driver.get_datasets().keys()))
    except exceptions.InvalidDatabaseError as exc:
        raise RuntimeError(
            "Error while connecting to remote database. Please double-check your AWS environment "
            "variables, and make sure your machine has access to the remote Terracotta database."
        ) from exc

    # this makes sure that a random raster file is readable
    with driver.connect():
        driver.get_raster_tile(some_dataset)



import pytest

import numpy as np


@pytest.fixture(autouse=True)
def reload_cmap():
    import importlib
    import terracotta.cmaps.get_cmaps

    try:
        yield
    finally:
        importlib.reload(terracotta.cmaps.get_cmaps)


def test_get_cmap():
    from terracotta.cmaps.get_cmaps import get_cmap, AVAILABLE_CMAPS

    for name in AVAILABLE_CMAPS:
        cmap = get_cmap(name)
        assert cmap.shape == (255, 4)
        assert cmap.dtype == np.uint8


def test_get_cmap_filesystem(monkeypatch):
    import pkg_resources
    import importlib

    import terracotta.cmaps.get_cmaps

    def throw_error(*args, **kwargs):
        raise pkg_resources.DistributionNotFound("monkeypatched")

    with monkeypatch.context() as m:
        m.setattr(pkg_resources.Requirement, "parse", throw_error)

        with pytest.raises(pkg_resources.DistributionNotFound):
            pkg_resources.Requirement.parse("terracotta")

        importlib.reload(terracotta.cmaps.get_cmaps)

        cmap = terracotta.cmaps.get_cmaps.get_cmap("jet")
        assert cmap.shape == (255, 4)
        assert cmap.dtype == np.uint8


def test_extra_cmap(monkeypatch, tmpdir):
    import importlib

    import terracotta.cmaps.get_cmaps

    custom_cmap_data = np.tile(np.arange(255, dtype="uint8"), (4, 1)).T
    np.save(str(tmpdir / f"foo{terracotta.cmaps.get_cmaps.SUFFIX}"), custom_cmap_data)
    np.save(str(tmpdir / "bar.npy"), custom_cmap_data)

    with monkeypatch.context() as m:
        m.setenv("TC_EXTRA_CMAP_FOLDER", str(tmpdir))
        importlib.reload(terracotta.cmaps.get_cmaps)

        assert "foo" in terracotta.cmaps.get_cmaps.AVAILABLE_CMAPS
        assert "bar" not in terracotta.cmaps.get_cmaps.AVAILABLE_CMAPS

        np.testing.assert_equal(
            custom_cmap_data, terracotta.cmaps.get_cmaps.get_cmap("foo")
        )


def test_extra_cmap_invalid_shape(monkeypatch, tmpdir):
    import importlib
    import terracotta.cmaps.get_cmaps

    broken_cmap_data = np.tile(np.arange(666, dtype="uint8"), (4, 1)).T
    np.save(str(tmpdir / f"foo{terracotta.cmaps.get_cmaps.SUFFIX}"), broken_cmap_data)

    with monkeypatch.context() as m:
        m.setenv("TC_EXTRA_CMAP_FOLDER", str(tmpdir))

        with pytest.raises(ValueError) as raised_exc:
            importlib.reload(terracotta.cmaps.get_cmaps)

        assert "foo" in str(raised_exc.value)
        assert "666" in str(raised_exc.value)


def test_extra_cmap_invalid_folder(monkeypatch):
    import importlib
    import terracotta.cmaps.get_cmaps

    with monkeypatch.context() as m:
        m.setenv("TC_EXTRA_CMAP_FOLDER", "bar")

        with pytest.raises(IOError) as raised_exc:
            importlib.reload(terracotta.cmaps.get_cmaps)

        assert "bar" in str(raised_exc.value)


def test_extra_cmap_invalid_dtype(monkeypatch, tmpdir):
    import importlib
    import terracotta.cmaps.get_cmaps

    broken_cmap_data = np.tile(np.arange(255, dtype="float"), (4, 1)).T
    np.save(str(tmpdir / f"foo{terracotta.cmaps.get_cmaps.SUFFIX}"), broken_cmap_data)

    with monkeypatch.context() as m:
        m.setenv("TC_EXTRA_CMAP_FOLDER", str(tmpdir))

        with pytest.raises(ValueError) as raised_exc:
            importlib.reload(terracotta.cmaps.get_cmaps)

        assert "foo" in str(raised_exc.value)
        assert "float" in str(raised_exc.value)



import pytest

import platform
import time

import numpy as np

from terracotta import exceptions

DRIVERS = ["sqlite"]  # A single MetaStore suffices for testing the RasterStore
METADATA_KEYS = ("bounds", "range", "mean", "stdev", "percentiles", "metadata")


@pytest.fixture
def dynamic_non_writable_db(monkeypatch, driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file), skip_metadata=True)

    with monkeypatch.context() as m:
        m.setattr(db.meta_store, "_WRITABLE", False)
        yield

    db.delete(["some", "value"])


@pytest.mark.parametrize("provider", DRIVERS)
def test_insertion_and_retrieval(driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))

    data = db.get_datasets()
    assert list(data.keys()) == [("some", "value")]
    assert data[("some", "value")] == str(raster_file)

    metadata = db.get_metadata(("some", "value"))
    assert all(key in metadata for key in METADATA_KEYS)


@pytest.mark.parametrize("provider", DRIVERS)
def test_path_override(driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")
    key_value = ("some", "value")
    bogus_path = "foo"

    db.create(keys)
    db.insert(key_value, str(raster_file), override_path=bogus_path)
    assert db.get_datasets()[key_value] == bogus_path

    with pytest.raises(IOError) as exc:
        # overridden path doesn't exist
        db.get_raster_tile(key_value)
        assert bogus_path in exc.value


@pytest.mark.parametrize("provider", DRIVERS)
def test_where(driver_path, provider, raster_file):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))
    db.insert(["some", "other_value"], str(raster_file))
    db.insert({"some": "a", "keynames": "third_value"}, str(raster_file))

    data = db.get_datasets()
    assert len(data) == 3

    data = db.get_datasets(where=dict(some="some"))
    assert len(data) == 2

    data = db.get_datasets(where=dict(some="some", keynames="value"))
    assert list(data.keys()) == [("some", "value")]
    assert data[("some", "value")] == str(raster_file)

    data = db.get_datasets(where=dict(some="unknown"))
    assert data == {}

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.get_datasets(where=dict(unknown="foo"))
    assert "unrecognized keys" in str(exc.value)


@pytest.mark.parametrize("provider", DRIVERS)
def test_where_with_multiquery(driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))
    db.insert(["some", "other_value"], str(raster_file))
    db.insert({"some": "a", "keynames": "third_value"}, str(raster_file))

    data = db.get_datasets()
    assert len(data) == 3

    data = db.get_datasets(where=dict(some=["some"]))
    assert len(data) == 2

    data = db.get_datasets(where=dict(keynames=["value", "other_value"]))
    assert len(data) == 2

    data = db.get_datasets(where=dict(some="some", keynames=["value", "third_value"]))
    assert list(data.keys()) == [("some", "value")]
    assert data[("some", "value")] == str(raster_file)

    data = db.get_datasets(where=dict(some=["unknown"]))
    assert data == {}


@pytest.mark.parametrize("provider", DRIVERS)
def test_pagination(driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))
    db.insert(["some", "other_value"], str(raster_file))
    db.insert({"some": "a", "keynames": "third_value"}, str(raster_file))

    data = db.get_datasets()
    assert len(data) == 3

    data = db.get_datasets(limit=2)
    assert len(data) == 2

    data = db.get_datasets(limit=2, page=1)
    assert len(data) == 1

    data = db.get_datasets(where=dict(some="some"), limit=1, page=0)
    assert len(data) == 1


@pytest.mark.parametrize("provider", DRIVERS)
def test_lazy_loading(driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file), skip_metadata=False)
    db.insert(["some", "other_value"], str(raster_file), skip_metadata=True)

    datasets = db.get_datasets()
    assert len(datasets) == 2

    data1 = db.get_metadata(["some", "value"])
    data2 = db.get_metadata({"some": "some", "keynames": "other_value"})
    assert set(data1.keys()) == set(data2.keys())
    assert all(np.all(data1[k] == data2[k]) for k in data1.keys())


@pytest.mark.parametrize("provider", DRIVERS)
def test_non_writable_lazy_loading(driver_path, provider, dynamic_non_writable_db):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)

    with pytest.raises(exceptions.DatabaseNotWritableError):
        db.get_metadata(["some", "value"])


@pytest.mark.parametrize("provider", DRIVERS)
def test_precomputed_metadata(driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    metadata = db.compute_metadata(str(raster_file))

    db.create(keys)
    db.insert(["some", "value"], str(raster_file), metadata=metadata)
    db.insert(["some", "other_value"], str(raster_file))

    datasets = db.get_datasets()
    assert len(datasets) == 2

    data1 = db.get_metadata(["some", "value"])
    data2 = db.get_metadata({"some": "some", "keynames": "other_value"})
    assert list(data1.keys()) == list(data2.keys())
    assert all(np.all(data1[k] == data2[k]) for k in data1.keys())


@pytest.mark.parametrize("provider", DRIVERS)
def test_invalid_insertion(monkeypatch, driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("keyname",)

    db.create(keys)

    def throw(*args, **kwargs):
        raise NotImplementedError()

    with monkeypatch.context() as m:
        m.setattr(db, "compute_metadata", throw)

        db.insert(["bar"], str(raster_file), skip_metadata=True)

        with pytest.raises(NotImplementedError):
            db.insert(["foo"], str(raster_file), skip_metadata=False)

        datasets = db.get_datasets()

        assert ("bar",) in datasets
        assert ("foo",) not in datasets


@pytest.mark.parametrize("provider", DRIVERS)
def test_wrong_key_number(driver_path, provider, raster_file):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("keyname",)

    db.create(keys)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.get_metadata(["a", "b"])
    assert "wrong number of keys" in str(exc.value)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.insert(["a", "b"], "", skip_metadata=True)
    assert "wrong number of keys" in str(exc.value)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.delete(["a", "b"])
    assert "wrong number of keys" in str(exc.value)


@pytest.mark.parametrize("provider", DRIVERS)
def test_invalid_group_insertion(monkeypatch, driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("keyname",)

    db.create(keys)

    def throw(*args, **kwargs):
        raise NotImplementedError()

    with monkeypatch.context() as m:
        m.setattr(db, "compute_metadata", throw)

        with db.connect():
            db.insert(["bar"], str(raster_file), skip_metadata=True)

            with pytest.raises(NotImplementedError):
                db.insert(["foo"], str(raster_file), skip_metadata=False)

            datasets = db.get_datasets()

        # first insertion succeeded, second failed
        assert ("bar",) in datasets
        assert ("foo",) not in datasets


def insertion_worker(key, path, raster_file, provider):
    import time
    from terracotta import drivers

    db = drivers.get_driver(path, provider=provider)
    with db.connect():
        db.insert([key], str(raster_file), skip_metadata=True)
        # keep connection open for a while to increase the chance of
        # triggering a race condition
        time.sleep(0.01)


@pytest.mark.xfail(platform.system() == "Darwin", reason="Flaky on OSX")
@pytest.mark.parametrize("provider", DRIVERS)
def test_multiprocess_insertion(driver_path, provider, raster_file):
    import functools
    import concurrent.futures
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    raster_file = str(raster_file)
    keys = ("keyname",)

    db.create(keys)

    key_vals = [str(i) for i in range(100)]

    worker = functools.partial(
        insertion_worker, path=driver_path, raster_file=raster_file, provider=provider
    )

    with concurrent.futures.ProcessPoolExecutor(4) as executor:
        for _ in executor.map(worker, key_vals):
            pass

    datasets = db.get_datasets()
    assert all((key,) in datasets for key in key_vals)

    data1 = db.get_metadata(["77"])
    data2 = db.get_metadata({"keyname": "99"})
    assert list(data1.keys()) == list(data2.keys())
    assert all(np.all(data1[k] == data2[k]) for k in data1.keys())


@pytest.mark.parametrize("provider", DRIVERS)
def test_insertion_invalid_raster(driver_path, provider, invalid_raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("keyname",)

    db.create(keys)

    with pytest.raises(ValueError):
        db.insert(["val"], str(invalid_raster_file))

    datasets = db.get_datasets()
    assert ("val",) not in datasets


@pytest.mark.parametrize("provider", DRIVERS)
@pytest.mark.parametrize("resampling_method", ["nearest", "linear", "cubic", "average"])
def test_raster_retrieval(driver_path, provider, raster_file, resampling_method):
    import terracotta

    terracotta.update_settings(RESAMPLING_METHOD=resampling_method)

    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))
    db.insert(["some", "other_value"], str(raster_file))

    data1 = db.get_raster_tile(["some", "value"], tile_size=(256, 256))
    assert data1.shape == (256, 256)

    data2 = db.get_raster_tile(["some", "other_value"], tile_size=(256, 256))
    assert data2.shape == (256, 256)

    np.testing.assert_array_equal(data1, data2)


@pytest.mark.parametrize("provider", DRIVERS)
@pytest.mark.parametrize("asynchronous", [True, False])
def test_raster_cache(driver_path, provider, raster_file, asynchronous):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))
    db.insert(["some", "other_value"], str(raster_file))

    assert len(db.raster_store._raster_cache) == 0

    data1 = db.get_raster_tile(
        ["some", "value"], tile_size=(256, 256), asynchronous=asynchronous
    )

    if asynchronous:
        data1 = data1.result()
        time.sleep(1)  # allow callback to finish

    assert len(db.raster_store._raster_cache) == 1

    data2 = db.get_raster_tile(
        ["some", "value"], tile_size=(256, 256), asynchronous=asynchronous
    )

    if asynchronous:
        data2 = data2.result()

    np.testing.assert_array_equal(data1, data2)
    assert len(db.raster_store._raster_cache) == 1


@pytest.mark.parametrize("provider", DRIVERS)
@pytest.mark.parametrize("asynchronous", [True, False])
def test_raster_cache_fail(driver_path, provider, raster_file, asynchronous):
    """Retrieve a tile that is larger than the total cache size"""
    from terracotta import drivers, update_settings

    update_settings(RASTER_CACHE_SIZE=1)

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))

    assert len(db.raster_store._raster_cache) == 0

    data1 = db.get_raster_tile(
        ["some", "value"], tile_size=(256, 256), asynchronous=asynchronous
    )

    if asynchronous:
        data1 = data1.result()
        time.sleep(1)  # allow callback to finish

    assert len(db.raster_store._raster_cache) == 0


@pytest.mark.parametrize("provider", DRIVERS)
def test_multiprocessing_fallback(driver_path, provider, raster_file, monkeypatch):
    import concurrent.futures
    from importlib import reload
    from terracotta import drivers
    import terracotta.drivers.geotiff_raster_store

    def dummy(*args, **kwargs):
        raise OSError("monkeypatched")

    try:
        with monkeypatch.context() as m, pytest.warns(UserWarning):
            m.setattr(concurrent.futures, "ProcessPoolExecutor", dummy)

            reload(terracotta.drivers.geotiff_raster_store)
            db = drivers.get_driver(driver_path, provider=provider)
            keys = ("some", "keynames")

            db.create(keys)
            db.insert(["some", "value"], str(raster_file))
            db.insert(["some", "other_value"], str(raster_file))

            data1 = db.get_raster_tile(["some", "value"], tile_size=(256, 256))
            assert data1.shape == (256, 256)

            data2 = db.get_raster_tile(["some", "other_value"], tile_size=(256, 256))
            assert data2.shape == (256, 256)

            np.testing.assert_array_equal(data1, data2)
    finally:
        reload(terracotta.drivers.geotiff_raster_store)


@pytest.mark.parametrize("provider", DRIVERS)
def test_raster_duplicate(driver_path, provider, raster_file):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))
    db.insert(["some", "value"], str(raster_file))

    assert list(db.get_datasets().keys()) == [("some", "value")]


@pytest.mark.parametrize("provider", DRIVERS)
def test_deletion(driver_path, provider, raster_file):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)

    dataset = {"some": "some", "keynames": "value"}
    db.insert(dataset, str(raster_file))

    data = db.get_datasets()
    assert list(data.keys()) == [("some", "value")]
    assert data[("some", "value")] == str(raster_file)

    metadata = db.get_metadata(("some", "value"))
    assert all(key in metadata for key in METADATA_KEYS)

    db.delete(dataset)
    assert not db.get_datasets()

    with pytest.raises(exceptions.DatasetNotFoundError):
        db.get_metadata(dataset)


@pytest.mark.parametrize("provider", DRIVERS)
def test_delete_nonexisting(driver_path, provider, raster_file):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)

    dataset = {"some": "some", "keynames": "value"}

    with pytest.raises(exceptions.DatasetNotFoundError):
        db.delete(dataset)


@pytest.mark.parametrize("provider", DRIVERS)
def test_nodata_consistency(
    driver_path, provider, big_raster_file_mask, big_raster_file_nodata
):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("type",)

    db.create(keys)
    db.insert(["mask"], str(big_raster_file_mask), skip_metadata=True)
    db.insert(["nodata"], str(big_raster_file_nodata), skip_metadata=True)

    data_mask = db.get_raster_tile(["mask"])
    data_nodata = db.get_raster_tile(["nodata"])

    np.testing.assert_array_equal(data_mask.mask, data_nodata.mask)


@pytest.mark.parametrize("provider", DRIVERS)
def test_broken_process_pool(monkeypatch, driver_path, provider, raster_file):
    import concurrent.futures
    import terracotta.drivers.geotiff_raster_store
    from terracotta import drivers

    class BrokenPool:
        def submit(self, *args, **kwargs):
            raise concurrent.futures.process.BrokenProcessPool("monkeypatched")

    with monkeypatch.context() as m:
        m.setattr(terracotta.drivers.geotiff_raster_store, "_executor", BrokenPool())

        db = drivers.get_driver(driver_path, provider=provider)
        keys = ("some", "keynames")

        db.create(keys)
        db.insert(["some", "value"], str(raster_file))
        db.insert(["some", "other_value"], str(raster_file))

        data1 = db.get_raster_tile(["some", "value"], tile_size=(256, 256))
        assert data1.shape == (256, 256)

        data2 = db.get_raster_tile(["some", "other_value"], tile_size=(256, 256))
        assert data2.shape == (256, 256)

        np.testing.assert_array_equal(data1, data2)


def test_no_multiprocessing():
    import concurrent.futures
    from terracotta import update_settings
    from terracotta.drivers.geotiff_raster_store import create_executor

    update_settings(USE_MULTIPROCESSING=False)

    executor = create_executor()
    assert isinstance(executor, concurrent.futures.ThreadPoolExecutor)



from PIL import Image
import numpy as np


def test_compute_handler(use_testdb, raster_file_xyz):
    import terracotta
    from terracotta.handlers import compute

    settings = terracotta.get_settings()

    raw_img = compute.compute(
        "v1 + v2",
        ["val21", "x"],
        {"v1": "val22", "v2": "val23"},
        stretch_range=(0, 10000),
        tile_xyz=raster_file_xyz,
    )
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == settings.DEFAULT_TILE_SIZE


def test_compute_consistency(use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import compute
    from terracotta.image import to_uint8

    settings = terracotta.get_settings()

    raw_img = compute.compute(
        "v1 + v2",
        ["val21", "x"],
        {"v1": "val22", "v2": "val23"},
        stretch_range=(0, 10000),
        tile_xyz=raster_file_xyz,
    )
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == settings.DEFAULT_TILE_SIZE

    driver = terracotta.get_driver(testdb)

    with driver.connect():
        v1 = get_tile_data(driver, ["val21", "x", "val22"], raster_file_xyz)
        v2 = get_tile_data(driver, ["val21", "x", "val23"], raster_file_xyz)

    np.testing.assert_array_equal(img_data, to_uint8(v1 + v2, 0, 10000))


def test_compute_transparency_nan(use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import compute

    raw_img = compute.compute(
        "where(v1 > 0, nan, v1 + v2)",
        ["val21", "x"],
        {"v1": "val22", "v2": "val23"},
        stretch_range=(0, 10000),
        tile_xyz=raster_file_xyz,
    )
    img_data = np.asarray(Image.open(raw_img).convert("RGBA"))

    driver = terracotta.get_driver(testdb)

    with driver.connect():
        v1 = get_tile_data(driver, ["val21", "x", "val22"], raster_file_xyz)

    alpha = img_data[..., 3]
    np.testing.assert_array_equal(
        alpha,
        np.where(v1.mask | (v1 > 0), 0, 255),
    )


def test_compute_transparency_mask(use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import compute

    raw_img = compute.compute(
        "masked_where(v1 > 0, v1 + v2)",
        ["val21", "x"],
        {"v1": "val22", "v2": "val23"},
        stretch_range=(0, 10000),
        tile_xyz=raster_file_xyz,
    )
    img_data = np.asarray(Image.open(raw_img).convert("RGBA"))

    driver = terracotta.get_driver(testdb)

    with driver.connect():
        v1 = get_tile_data(driver, ["val21", "x", "val22"], raster_file_xyz)

    alpha = img_data[..., 3]
    np.testing.assert_array_equal(
        alpha,
        np.where(v1.mask | (v1 > 0), 0, 255),
    )



"""server/singleband.py

Flask route to handle /singleband calls.
"""

from typing import Optional, Any, Mapping, Dict, Tuple
import json

from marshmallow import (
    Schema,
    fields,
    validate,
    validates_schema,
    pre_load,
    ValidationError,
    EXCLUDE,
)
from flask import request, send_file, Response

from terracotta.server.fields import StringOrNumber, validate_stretch_range
from terracotta.server.flask_api import TILE_API
from terracotta.cmaps import AVAILABLE_CMAPS


class SinglebandQuerySchema(Schema):
    keys = fields.String(
        required=True, description="Keys identifying dataset, in order"
    )
    tile_z = fields.Int(required=True, description="Requested zoom level")
    tile_y = fields.Int(required=True, description="y coordinate")
    tile_x = fields.Int(required=True, description="x coordinate")


class SinglebandOptionSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    stretch_range = fields.List(
        StringOrNumber(allow_none=True, validate=validate_stretch_range),
        validate=validate.Length(equal=2),
        example="[0,1]",
        description=(
            "Stretch range [min, max] to use as JSON array. "
            "Min and max may be numbers to use as absolute range, or strings "
            "of the format `p<digits>` with an integer between 0 and 100 "
            "to use percentiles of the image instead. "
            "Null values indicate global minimum / maximum."
        ),
        missing=None,
    )

    colormap = fields.String(
        description="Colormap to apply to image (see /colormap)",
        validate=validate.OneOf(("explicit", *AVAILABLE_CMAPS)),
        missing=None,
    )

    explicit_color_map = fields.Dict(
        keys=fields.Number(),
        values=fields.List(fields.Number, validate=validate.Length(min=3, max=4)),
        example='{"0": [255, 255, 255]}',
        description="Explicit value-color mapping to use, encoded as JSON object. "
        "Must be given together with `colormap=explicit`. Color values can be "
        "specified either as RGB or RGBA tuple (in the range of [0, 255]), or as "
        "hex strings.",
    )

    tile_size = fields.List(
        fields.Integer(),
        validate=validate.Length(equal=2),
        example="[256,256]",
        description="Pixel dimensions of the returned PNG image as JSON list.",
    )

    @validates_schema
    def validate_cmap(self, data: Mapping[str, Any], **kwargs: Any) -> None:
        if data.get("colormap", "") == "explicit" and not data.get(
            "explicit_color_map"
        ):
            raise ValidationError(
                "explicit_color_map argument must be given for colormap=explicit",
                "colormap",
            )

        if data.get("explicit_color_map") and data.get("colormap", "") != "explicit":
            raise ValidationError(
                "explicit_color_map can only be given for colormap=explicit",
                "explicit_color_map",
            )

    @pre_load
    def decode_json(self, data: Mapping[str, Any], **kwargs: Any) -> Dict[str, Any]:
        data = dict(data.items())
        for var in ("stretch_range", "tile_size", "explicit_color_map"):
            val = data.get(var)
            if val:
                try:
                    data[var] = json.loads(val)
                except json.decoder.JSONDecodeError as exc:
                    msg = f"Could not decode value {val} for {var} as JSON"
                    raise ValidationError(msg) from exc

        val = data.get("explicit_color_map")
        if val and isinstance(val, dict):
            for key, color in val.items():
                if isinstance(color, str):
                    # convert hex strings to RGBA
                    hex_string = color.lstrip("#")
                    try:
                        rgb = [int(hex_string[i : i + 2], 16) for i in (0, 2, 4)]
                        data["explicit_color_map"][key] = (*rgb, 255)
                    except ValueError:
                        msg = f"Could not decode value {color} in explicit_color_map as hex string"
                        raise ValidationError(msg)
                elif len(color) == 3:
                    # convert RGB to RGBA
                    data["explicit_color_map"][key] = (*color, 255)

        return data


@TILE_API.route(
    "/singleband/<path:keys>/<int:tile_z>/<int:tile_x>/<int:tile_y>.png",
    methods=["GET"],
)
def get_singleband(tile_z: int, tile_y: int, tile_x: int, keys: str) -> Response:
    """Return single-band PNG image of requested tile
    ---
    get:
        summary: /singleband (tile)
        description: Return single-band PNG image of requested XYZ tile
        parameters:
            - in: path
              schema: SinglebandQuerySchema
            - in: query
              schema: SinglebandOptionSchema
        responses:
            200:
                description:
                    PNG image of requested tile
            400:
                description:
                    Invalid query parameters
            404:
                description:
                    No dataset found for given key combination
    """
    tile_xyz = (tile_x, tile_y, tile_z)
    return _get_singleband_image(keys, tile_xyz)


class SinglebandPreviewSchema(Schema):
    keys = fields.String(
        required=True, description="Keys identifying dataset, in order"
    )


@TILE_API.route("/singleband/<path:keys>/preview.png", methods=["GET"])
def get_singleband_preview(keys: str) -> Response:
    """Return single-band PNG preview image of requested dataset
    ---
    get:
        summary: /singleband (preview)
        description: Return single-band PNG preview image of requested dataset
        parameters:
            - in: path
              schema: SinglebandPreviewSchema
            - in: query
              schema: SinglebandOptionSchema
        responses:
            200:
                description:
                    PNG image of requested tile
            400:
                description:
                    Invalid query parameters
            404:
                description:
                    No dataset found for given key combination
    """
    return _get_singleband_image(keys)


def _get_singleband_image(
    keys: str, tile_xyz: Optional[Tuple[int, int, int]] = None
) -> Response:
    from terracotta.handlers.singleband import singleband

    parsed_keys = [key for key in keys.split("/") if key]

    option_schema = SinglebandOptionSchema()
    options = option_schema.load(request.args)

    if options.get("colormap", "") == "explicit":
        options["colormap"] = options.pop("explicit_color_map")

    image = singleband(parsed_keys, tile_xyz=tile_xyz, **options)

    return send_file(image, mimetype="image/png")



"""drivers/geotiff_raster_store.py

Base class for drivers operating on physical raster files.
"""

from typing import Optional, Any, Callable, Sequence, Dict, TypeVar
from concurrent.futures import Future, Executor, ProcessPoolExecutor, ThreadPoolExecutor
from concurrent.futures.process import BrokenProcessPool

import functools
import logging
import warnings
import threading

import numpy as np

from terracotta import get_settings
from terracotta import raster
from terracotta.cache import CompressedLFUCache
from terracotta.drivers.base_classes import RasterStore

Number = TypeVar("Number", int, float)

logger = logging.getLogger(__name__)

_executor = None


def create_executor() -> Executor:
    settings = get_settings()

    if not settings.USE_MULTIPROCESSING:
        return ThreadPoolExecutor(max_workers=1)

    executor: Executor

    try:
        # this fails on architectures without /dev/shm
        executor = ProcessPoolExecutor(max_workers=3)
    except OSError:
        # fall back to serial evaluation
        warnings.warn(
            "Multiprocessing is not available on this system. "
            "Falling back to serial execution."
        )
        executor = ThreadPoolExecutor(max_workers=1)

    return executor


def submit_to_executor(task: Callable[..., Any]) -> Future:
    global _executor

    if _executor is None:
        _executor = create_executor()

    try:
        future = _executor.submit(task)
    except BrokenProcessPool:
        # re-create executor and try again
        logger.warn("Re-creating broken process pool")
        _executor = create_executor()
        future = _executor.submit(task)

    return future


def ensure_hashable(val: Any) -> Any:
    if isinstance(val, list):
        return tuple(val)

    if isinstance(val, dict):
        return tuple((k, ensure_hashable(v)) for k, v in val.items())

    return val


class GeoTiffRasterStore(RasterStore):
    """Raster store that operates on GeoTiff raster files from disk.

    Path arguments are expected to be file paths.
    """

    _TARGET_CRS: str = "epsg:3857"
    _LARGE_RASTER_THRESHOLD: int = 10980 * 10980
    _RIO_ENV_OPTIONS = dict(
        GDAL_TIFF_INTERNAL_MASK=True, GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR"
    )

    def __init__(self) -> None:
        settings = get_settings()
        self._raster_cache = CompressedLFUCache(
            settings.RASTER_CACHE_SIZE,
            compression_level=settings.RASTER_CACHE_COMPRESS_LEVEL,
        )
        self._cache_lock = threading.RLock()

    def compute_metadata(
        self,
        path: str,
        *,
        extra_metadata: Optional[Any] = None,
        use_chunks: Optional[bool] = None,
        max_shape: Optional[Sequence[int]] = None
    ) -> Dict[str, Any]:
        return raster.compute_metadata(
            path,
            extra_metadata=extra_metadata,
            use_chunks=use_chunks,
            max_shape=max_shape,
            large_raster_threshold=self._LARGE_RASTER_THRESHOLD,
            rio_env_options=self._RIO_ENV_OPTIONS,
        )

    # return type has to be Any until mypy supports conditional return types
    def get_raster_tile(
        self,
        path: str,
        *,
        tile_bounds: Optional[Sequence[float]] = None,
        tile_size: Optional[Sequence[int]] = None,
        preserve_values: bool = False,
        asynchronous: bool = False
    ) -> Any:
        future: Future[np.ma.MaskedArray]
        result: np.ma.MaskedArray

        settings = get_settings()

        if tile_size is None:
            tile_size = settings.DEFAULT_TILE_SIZE

        kwargs = dict(
            path=path,
            tile_bounds=tile_bounds,
            tile_size=tuple(tile_size),
            preserve_values=preserve_values,
            reprojection_method=settings.REPROJECTION_METHOD,
            resampling_method=settings.RESAMPLING_METHOD,
            target_crs=self._TARGET_CRS,
            rio_env_options=self._RIO_ENV_OPTIONS,
        )

        cache_key = hash(ensure_hashable(kwargs))

        try:
            with self._cache_lock:
                result = self._raster_cache[cache_key]
        except KeyError:
            pass
        else:
            if asynchronous:
                # wrap result in a future
                future = Future()
                future.set_result(result)
                return future
            else:
                return result

        retrieve_tile = functools.partial(raster.get_raster_tile, **kwargs)

        future = submit_to_executor(retrieve_tile)

        def cache_callback(future: Future) -> None:
            # insert result into global cache if execution was successful
            if future.exception() is None:
                self._add_to_cache(cache_key, future.result())

        if asynchronous:
            future.add_done_callback(cache_callback)
            return future
        else:
            result = future.result()
            cache_callback(future)
            return result

    def _add_to_cache(self, key: Any, value: Any) -> None:
        try:
            with self._cache_lock:
                self._raster_cache[key] = value
        except ValueError:  # value too large
            pass



"""drivers/relational_meta_store.py

Base class for relational database drivers, using SQLAlchemy.
"""

import contextlib
import functools
import json
import re
import urllib.parse as urlparse
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Iterator, Mapping, Optional, Sequence, Tuple, Type, Union

import numpy as np
import sqlalchemy as sqla
import terracotta
from sqlalchemy.engine.base import Connection
from sqlalchemy.engine.url import URL
from terracotta import exceptions
from terracotta.drivers.base_classes import (
    KeysType,
    MetaStore,
    MultiValueKeysType,
    requires_writable,
)
from terracotta.profile import trace

_ERROR_ON_CONNECT = (
    "Could not connect to database. Make sure that the given path points "
    "to a valid Terracotta database, and that you ran driver.create()."
)

DATABASE_DRIVER_EXCEPTIONS_TO_CONVERT: Tuple[Type[Exception], ...] = (
    sqla.exc.OperationalError,
    sqla.exc.InternalError,
    sqla.exc.ProgrammingError,
    sqla.exc.InvalidRequestError,
)

ExceptionType = Union[Type[Exception], Tuple[Type[Exception], ...]]


@contextlib.contextmanager
def convert_exceptions(
    error_message: str,
    exceptions_to_convert: ExceptionType = DATABASE_DRIVER_EXCEPTIONS_TO_CONVERT,
) -> Iterator:
    try:
        yield
    except exceptions_to_convert as exception:
        raise exceptions.InvalidDatabaseError(error_message) from exception


@contextlib.contextmanager
def try_database_operation(connection: Connection) -> Iterator:
    try:
        yield
    except:  # noqa: E722
        connection.rollback()
        raise
    else:
        connection.commit()


class RelationalMetaStore(MetaStore, ABC):
    SQL_DIALECT: str  # The database flavour, eg mysql, sqlite, etc
    SQL_DRIVER: str  # The actual database driver, eg pymysql, sqlite3, etc
    SQL_KEY_SIZE: int
    SQL_TIMEOUT_KEY: str

    SQLA_STRING: Any = sqla.types.String
    SQLA_METADATA_TYPE_LOOKUP: Dict[str, Any] = {
        "real": functools.partial(sqla.types.Float, precision=8),
        "text": sqla.types.Text,
        "blob": sqla.types.LargeBinary,
    }

    _METADATA_COLUMNS: Tuple[Tuple[str, str], ...] = (
        ("bounds_north", "real"),
        ("bounds_east", "real"),
        ("bounds_south", "real"),
        ("bounds_west", "real"),
        ("convex_hull", "text"),
        ("valid_percentage", "real"),
        ("min", "real"),
        ("max", "real"),
        ("mean", "real"),
        ("stdev", "real"),
        ("percentiles", "blob"),
        ("metadata", "text"),
    )

    def __init__(self, path: str) -> None:
        settings = terracotta.get_settings()
        db_connection_timeout: int = settings.DB_CONNECTION_TIMEOUT

        self.url = self._parse_path(path)
        self.sqla_engine = sqla.create_engine(
            self.url,
            echo=False,
            future=True,
            connect_args={self.SQL_TIMEOUT_KEY: db_connection_timeout},
            # automatically re-spawn stale connections, see terracotta#266
            pool_pre_ping=True,
        )
        self.sqla_metadata = sqla.MetaData()

        self._db_keys: Optional[OrderedDict] = None

        self._connection: Optional[Connection] = None
        self.connected: bool = False
        self.db_version_verified: bool = False

        # use normalized path to make sure username and password don't leak into __repr__
        super().__init__(self._normalize_path(path))

    @classmethod
    def _parse_path(cls, connection_string: str) -> URL:
        if "//" not in connection_string:
            connection_string = f"//{connection_string}"

        con_params = urlparse.urlparse(connection_string)

        if not con_params.scheme:
            con_params = urlparse.urlparse(f"{cls.SQL_DIALECT}:{connection_string}")

        if con_params.scheme != cls.SQL_DIALECT:
            raise ValueError(f'unsupported URL scheme "{con_params.scheme}"')

        settings = terracotta.get_settings()
        url = URL.create(
            drivername=f"{cls.SQL_DIALECT}+{cls.SQL_DRIVER}",
            username=con_params.username or settings.SQL_USER,
            password=con_params.password or settings.SQL_PASSWORD,
            host=con_params.hostname,
            port=con_params.port,
            database=con_params.path[1:],  # remove leading '/' from urlparse
            query=dict(urlparse.parse_qsl(con_params.query)),
        )

        return url

    @contextlib.contextmanager
    def connect(self, verify: bool = True) -> Iterator:
        @convert_exceptions(_ERROR_ON_CONNECT, sqla.exc.OperationalError)
        def get_connection() -> Connection:
            return self.sqla_engine.connect().execution_options(
                isolation_level="READ UNCOMMITTED"
            )

        if not self.connected:
            # start a new connection
            try:
                with get_connection() as connection:
                    self._connection = connection
                    self.connected = True
                    if verify:
                        self._connection_callback()

                    with try_database_operation(self._connection):
                        yield self._connection
            finally:
                self._connection = None
                self.connected = False
        else:
            # re-use existing connection
            assert self._connection is not None
            with try_database_operation(self._connection):
                yield self._connection

    def _connection_callback(self) -> None:
        if not self.db_version_verified:
            # check for version compatibility
            def version_tuple(version_string: str) -> Sequence[str]:
                return version_string.split(".")

            db_version = self.db_version
            current_version = terracotta.__version__

            if version_tuple(db_version)[:2] != version_tuple(current_version)[:2]:
                raise exceptions.InvalidDatabaseError(
                    f"Version conflict: database was created in v{db_version}, "
                    f"but this is v{current_version}"
                )
            self.db_version_verified = True

    @property
    @convert_exceptions(_ERROR_ON_CONNECT)
    def db_version(self) -> str:
        """Terracotta version used to create the database"""
        terracotta_table = sqla.Table(
            "terracotta", self.sqla_metadata, autoload_with=self.sqla_engine
        )
        stmt = sqla.select(terracotta_table.c.version)

        with self.connect() as conn:
            version = conn.execute(stmt).scalar()

        return str(version)

    @requires_writable
    @convert_exceptions("Could not create database")
    def create(
        self, keys: Sequence[str], key_descriptions: Optional[Mapping[str, str]] = None
    ) -> None:
        """Create and initialize database with empty tables.

        This must be called before opening the first connection. The database must not
        exist already.

        Arguments:

            keys: Key names to use throughout the Terracotta database.
            key_descriptions: Optional (but recommended) full-text description for some keys,
                in the form of ``{key_name: description}``.

        """
        self._create_database()
        self._initialize_database(keys, key_descriptions)

    @abstractmethod
    def _create_database(self) -> None:
        # Note that some subclasses may not actually create any database here, as
        # it may be created automatically on connection for some database vendors
        pass

    def _initialize_database(
        self, keys: Sequence[str], key_descriptions: Optional[Mapping[str, str]] = None
    ) -> None:
        if key_descriptions is None:
            key_descriptions = {}
        else:
            key_descriptions = dict(key_descriptions)

        if not all(k in keys for k in key_descriptions.keys()):
            raise exceptions.InvalidKeyError(
                "key description dict contains unknown keys"
            )

        if not all(re.match(r"^\w+$", key) for key in keys):
            raise exceptions.InvalidKeyError("key names must be alphanumeric")

        if any(key in self._RESERVED_KEYS for key in keys):
            raise exceptions.InvalidKeyError(
                f"key names cannot be one of {self._RESERVED_KEYS!s}"
            )

        terracotta_table = sqla.Table(
            "terracotta",
            self.sqla_metadata,
            sqla.Column("version", self.SQLA_STRING(255), primary_key=True),
        )
        key_names_table = sqla.Table(
            "key_names",
            self.sqla_metadata,
            sqla.Column(
                "key_name", self.SQLA_STRING(self.SQL_KEY_SIZE), primary_key=True
            ),
            sqla.Column("description", self.SQLA_STRING(8000)),
            sqla.Column("idx", sqla.types.Integer, unique=True),
        )
        _ = sqla.Table(
            "datasets",
            self.sqla_metadata,
            *[
                sqla.Column(key, self.SQLA_STRING(self.SQL_KEY_SIZE), primary_key=True)
                for key in keys
            ],
            sqla.Column("path", self.SQLA_STRING(8000)),
        )
        _ = sqla.Table(
            "metadata",
            self.sqla_metadata,
            *[
                sqla.Column(key, self.SQLA_STRING(self.SQL_KEY_SIZE), primary_key=True)
                for key in keys
            ],
            *[
                sqla.Column(name, self.SQLA_METADATA_TYPE_LOOKUP[column_type]())
                for name, column_type in self._METADATA_COLUMNS
            ],
        )
        self.sqla_metadata.create_all(self.sqla_engine)

        with self.connect(verify=False) as conn:
            conn.execute(
                terracotta_table.insert().values(version=terracotta.__version__)
            )
            conn.execute(
                key_names_table.insert(),
                [
                    dict(key_name=key, description=key_descriptions.get(key, ""), idx=i)
                    for i, key in enumerate(keys)
                ],
            )

    @convert_exceptions("Could not retrieve keys from database")
    def get_keys(self) -> OrderedDict:
        keys_table = sqla.Table(
            "key_names", self.sqla_metadata, autoload_with=self.sqla_engine
        )

        with self.connect() as conn:
            result = conn.execute(
                sqla.select(
                    keys_table.c["key_name"], keys_table.c["description"]
                ).order_by(keys_table.c["idx"])
            )
        return OrderedDict((row.key_name, row.description) for row in result.all())

    @property
    def key_names(self) -> Tuple[str, ...]:
        """Names of all keys defined by the database"""
        if self._db_keys is None:
            self._db_keys = self.get_keys()
        return tuple(self._db_keys.keys())

    @trace("get_datasets")
    @convert_exceptions("Could not retrieve datasets")
    def get_datasets(
        self,
        where: Optional[MultiValueKeysType] = None,
        page: int = 0,
        limit: Optional[int] = None,
    ) -> Dict[Tuple[str, ...], str]:
        if where is None:
            where = {}

        where = {
            key: value if isinstance(value, list) else [value]
            for key, value in where.items()
        }

        datasets_table = sqla.Table(
            "datasets", self.sqla_metadata, autoload_with=self.sqla_engine
        )
        stmt = (
            datasets_table.select()
            .where(
                *[
                    sqla.or_(*[datasets_table.c[column] == value for value in values])
                    for column, values in where.items()
                ]
            )
            .order_by(*datasets_table.c.values())
        )

        if limit is not None:
            stmt = stmt.limit(limit).offset(page * limit if limit is not None else None)

        with self.connect() as conn:
            result = conn.execute(stmt).all()

        def keytuple(row: sqla.engine.row.Row) -> Tuple[str, ...]:
            return tuple(getattr(row, key) for key in self.key_names)

        datasets = {keytuple(row): row.path for row in result}
        return datasets

    @trace("get_metadata")
    @convert_exceptions("Could not retrieve metadata")
    def get_metadata(self, keys: KeysType) -> Optional[Dict[str, Any]]:
        metadata_table = sqla.Table(
            "metadata", self.sqla_metadata, autoload_with=self.sqla_engine
        )
        stmt = metadata_table.select().where(
            *[metadata_table.c[key] == value for key, value in keys.items()]
        )

        with self.connect() as conn:
            row = conn.execute(stmt).first()
        if not row:
            return None

        data_columns, _ = zip(*self._METADATA_COLUMNS)
        encoded_data = {col: getattr(row, col) for col in self.key_names + data_columns}
        return self._decode_data(encoded_data)

    @trace("insert")
    @requires_writable
    @convert_exceptions("Could not write to database")
    def insert(
        self, keys: KeysType, path: str, *, metadata: Optional[Mapping[str, Any]] = None
    ) -> None:
        datasets_table = sqla.Table(
            "datasets", self.sqla_metadata, autoload_with=self.sqla_engine
        )
        metadata_table = sqla.Table(
            "metadata", self.sqla_metadata, autoload_with=self.sqla_engine
        )

        with self.connect() as conn:
            conn.execute(
                datasets_table.delete().where(
                    *[
                        datasets_table.c[column] == value
                        for column, value in keys.items()
                    ]
                )
            )
            conn.execute(datasets_table.insert().values(**keys, path=path))

            if metadata is not None:
                encoded_data = self._encode_data(metadata)
                conn.execute(
                    metadata_table.delete().where(
                        *[
                            metadata_table.c[column] == value
                            for column, value in keys.items()
                        ]
                    )
                )
                conn.execute(metadata_table.insert().values(**keys, **encoded_data))

    @trace("delete")
    @requires_writable
    @convert_exceptions("Could not write to database")
    def delete(self, keys: KeysType) -> None:
        if not self.get_datasets(keys):
            raise exceptions.DatasetNotFoundError(f"No dataset found with keys {keys}")

        datasets_table = sqla.Table(
            "datasets", self.sqla_metadata, autoload_with=self.sqla_engine
        )
        metadata_table = sqla.Table(
            "metadata", self.sqla_metadata, autoload_with=self.sqla_engine
        )

        with self.connect() as conn:
            conn.execute(
                datasets_table.delete().where(
                    *[
                        datasets_table.c[column] == value
                        for column, value in keys.items()
                    ]
                )
            )
            conn.execute(
                metadata_table.delete().where(
                    *[
                        metadata_table.c[column] == value
                        for column, value in keys.items()
                    ]
                )
            )

    @staticmethod
    def _encode_data(decoded: Mapping[str, Any]) -> Dict[str, Any]:
        """Transform from internal format to database representation"""
        encoded = {
            "bounds_north": decoded["bounds"][0],
            "bounds_east": decoded["bounds"][1],
            "bounds_south": decoded["bounds"][2],
            "bounds_west": decoded["bounds"][3],
            "convex_hull": json.dumps(decoded["convex_hull"]),
            "valid_percentage": decoded["valid_percentage"],
            "min": decoded["range"][0],
            "max": decoded["range"][1],
            "mean": decoded["mean"],
            "stdev": decoded["stdev"],
            "percentiles": np.array(decoded["percentiles"], dtype="float32").tobytes(),
            "metadata": json.dumps(decoded["metadata"]),
        }
        return encoded

    @staticmethod
    def _decode_data(encoded: Mapping[str, Any]) -> Dict[str, Any]:
        """Transform from database format to internal representation"""
        decoded = {
            "bounds": tuple(
                [encoded[f"bounds_{d}"] for d in ("north", "east", "south", "west")]
            ),
            "convex_hull": json.loads(encoded["convex_hull"]),
            "valid_percentage": encoded["valid_percentage"],
            "range": (encoded["min"], encoded["max"]),
            "mean": encoded["mean"],
            "stdev": encoded["stdev"],
            "percentiles": np.frombuffer(
                encoded["percentiles"], dtype="float32"
            ).tolist(),
            "metadata": json.loads(encoded["metadata"]),
        }
        return decoded



"""cog.py

Provides a validator for cloud-optimized GeoTiff.
"""

from typing import Tuple, List, Dict, Any

import os

import rasterio
from rasterio.env import GDALVersion

ValidationInfo = Tuple[List[str], List[str], Dict[str, Any]]


def validate(src_path: str, strict: bool = True) -> bool:
    """Validate given cloud-optimized GeoTIFF"""
    errors, warnings, details = check_raster_file(src_path)
    if strict:
        return not errors and not warnings
    else:
        return not errors


def check_raster_file(src_path: str) -> ValidationInfo:  # pragma: no cover
    """
    Implementation from
    https://github.com/cogeotiff/rio-cogeo/blob/a07d914e2d898878417638bbc089179f01eb5b28/rio_cogeo/cogeo.py#L385

    This function is the rasterio equivalent of
    https://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/validate_cloud_optimized_geotiff.py
    """
    errors: List[str] = []
    warnings: List[str] = []
    details: Dict[str, Any] = {}

    if not GDALVersion.runtime().at_least("2.2"):
        raise RuntimeError("GDAL 2.2 or above required")

    config = dict(GDAL_DISABLE_READDIR_ON_OPEN="FALSE")
    with rasterio.Env(**config):
        with rasterio.open(src_path) as src:
            if not src.driver == "GTiff":
                errors.append("The file is not a GeoTIFF")
                return errors, warnings, details

            if any(os.path.splitext(x)[-1] == ".ovr" for x in src.files):
                errors.append(
                    "Overviews found in external .ovr file. They should be internal"
                )

            overviews = src.overviews(1)
            if src.width > 512 and src.height > 512:
                if not src.is_tiled:
                    errors.append(
                        "The file is greater than 512xH or 512xW, but is not tiled"
                    )

                if not overviews:
                    warnings.append(
                        "The file is greater than 512xH or 512xW, it is recommended "
                        "to include internal overviews"
                    )

            ifd_offset = int(src.get_tag_item("IFD_OFFSET", "TIFF", bidx=1))
            # Starting from GDAL 3.1, GeoTIFF and COG have ghost headers
            # e.g:
            # """
            # GDAL_STRUCTURAL_METADATA_SIZE=000140 bytes
            # LAYOUT=IFDS_BEFORE_DATA
            # BLOCK_ORDER=ROW_MAJOR
            # BLOCK_LEADER=SIZE_AS_UINT4
            # BLOCK_TRAILER=LAST_4_BYTES_REPEATED
            # KNOWN_INCOMPATIBLE_EDITION=NO
            # """
            #
            # This header should be < 200bytes
            if ifd_offset > 300:
                errors.append(
                    f"The offset of the main IFD should be < 300. It is {ifd_offset} instead"
                )

            ifd_offsets = [ifd_offset]
            details["ifd_offsets"] = {}
            details["ifd_offsets"]["main"] = ifd_offset

            if overviews and overviews != sorted(overviews):
                errors.append("Overviews should be sorted")

            for ix, dec in enumerate(overviews):

                # NOTE: Size check is handled in rasterio `src.overviews` methods
                # https://github.com/mapbox/rasterio/blob/4ebdaa08cdcc65b141ed3fe95cf8bbdd9117bc0b/rasterio/_base.pyx
                # We just need to make sure the decimation level is > 1
                if not dec > 1:
                    errors.append(
                        "Invalid Decimation {} for overview level {}".format(dec, ix)
                    )

                # Check that the IFD of descending overviews are sorted by increasing
                # offsets
                ifd_offset = int(src.get_tag_item("IFD_OFFSET", "TIFF", bidx=1, ovr=ix))
                ifd_offsets.append(ifd_offset)

                details["ifd_offsets"]["overview_{}".format(ix)] = ifd_offset
                if ifd_offsets[-1] < ifd_offsets[-2]:
                    if ix == 0:
                        errors.append(
                            "The offset of the IFD for overview of index {} is {}, "
                            "whereas it should be greater than the one of the main "
                            "image, which is at byte {}".format(
                                ix, ifd_offsets[-1], ifd_offsets[-2]
                            )
                        )
                    else:
                        errors.append(
                            "The offset of the IFD for overview of index {} is {}, "
                            "whereas it should be greater than the one of index {}, "
                            "which is at byte {}".format(
                                ix, ifd_offsets[-1], ix - 1, ifd_offsets[-2]
                            )
                        )

            block_offset = src.get_tag_item("BLOCK_OFFSET_0_0", "TIFF", bidx=1)

            data_offset = int(block_offset) if block_offset else 0
            data_offsets = [data_offset]
            details["data_offsets"] = {}
            details["data_offsets"]["main"] = data_offset

            for ix, dec in enumerate(overviews):
                block_offset = src.get_tag_item(
                    "BLOCK_OFFSET_0_0", "TIFF", bidx=1, ovr=ix
                )
                data_offset = int(block_offset) if block_offset else 0
                data_offsets.append(data_offset)
                details["data_offsets"]["overview_{}".format(ix)] = data_offset

            if data_offsets[-1] != 0 and data_offsets[-1] < ifd_offsets[-1]:
                if len(overviews) > 0:
                    errors.append(
                        "The offset of the first block of the smallest overview "
                        "should be after its IFD"
                    )
                else:
                    errors.append(
                        "The offset of the first block of the image should "
                        "be after its IFD"
                    )

            for i in range(len(data_offsets) - 2, 0, -1):
                if data_offsets[i] < data_offsets[i + 1]:
                    errors.append(
                        "The offset of the first block of overview of index {} should "
                        "be after the one of the overview of index {}".format(i - 1, i)
                    )

            if len(data_offsets) >= 2 and data_offsets[0] < data_offsets[1]:
                errors.append(
                    "The offset of the first block of the main resolution image "
                    "should be after the one of the overview of index {}".format(
                        len(overviews) - 1
                    )
                )

        for ix, dec in enumerate(overviews):
            with rasterio.open(src_path, OVERVIEW_LEVEL=ix) as ovr_dst:
                if ovr_dst.width > 512 and ovr_dst.height > 512:
                    if not ovr_dst.is_tiled:
                        errors.append("Overview of index {} is not tiled".format(ix))

    return errors, warnings, details



def test_key_handler(testdb, use_testdb):
    import terracotta
    from terracotta.handlers import keys

    driver = terracotta.get_driver(str(testdb))

    handler_response = keys.keys()
    assert handler_response
    assert tuple(row["key"] for row in handler_response) == driver.key_names


def test_key_handler_env_config(testdb, monkeypatch):
    import terracotta
    from terracotta.handlers import keys

    monkeypatch.setenv("TC_DRIVER_PATH", str(testdb))
    terracotta.update_settings()
    driver = terracotta.get_driver(str(testdb))

    handler_response = keys.keys()
    assert handler_response
    assert tuple(row["key"] for row in handler_response) == driver.key_names



"""server/spec.py

Flask routes to serve API spec
"""

from flask import jsonify, render_template, Response

from terracotta.server.flask_api import SPEC_API, SPEC


@SPEC_API.route("/swagger.json", methods=["GET"])
def specification() -> Response:
    return jsonify(SPEC.to_dict())


@SPEC_API.route("/apidoc", methods=["GET"])
def ui() -> str:
    return render_template("apidoc.html")



import pytest

TESTABLE_DRIVERS = ["sqlite", "mysql", "postgresql"]
DRIVER_CLASSES = {
    "sqlite": "SQLiteMetaStore",
    "sqlite-remote": "SQLiteRemoteMetaStore",
    "mysql": "MySQLMetaStore",
    "postgresql": "PostgreSQLMetaStore",
}


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_auto_detect(driver_path, provider):
    from terracotta import drivers

    db = drivers.get_driver(driver_path)
    assert db.meta_store.__class__.__name__ == DRIVER_CLASSES[provider]
    assert drivers.get_driver(driver_path, provider=provider) is db


def test_normalize_base(tmpdir):
    from terracotta.drivers import MetaStore

    # base class normalize is noop
    assert MetaStore._normalize_path(str(tmpdir)) == str(tmpdir)


@pytest.mark.parametrize("provider", ["sqlite"])
def test_normalize_filepath(provider, tmpdir):
    from terracotta.drivers import load_driver

    tmpdir = str(tmpdir)

    equivalent_paths = (
        f"{tmpdir}/foo.tc",
        f"{tmpdir}//foo.tc",
        f"{tmpdir}/bar/../foo.tc",
    )

    driver = load_driver(provider)
    first_path = driver._normalize_path(equivalent_paths[0])

    for p in equivalent_paths[1:]:
        assert driver._normalize_path(p) == first_path


@pytest.mark.parametrize("provider", ["mysql", "sqlite-remote"])
def test_normalize_url(provider):
    from terracotta.drivers import load_driver

    scheme = "mysql" if provider == "mysql" else "https"
    port = 3306 if provider == "mysql" else 443

    equivalent_urls = (
        "test.example.com/foo",
        "user@test.example.com/foo",
        "test.example.com/foo/",
        f"{scheme}://test.example.com/foo",
        f"{scheme}://user:password@test.example.com/foo",
        f"test.example.com:{port}/foo",
    )

    driver = load_driver(provider)
    first_path = driver._normalize_path(equivalent_urls[0])

    for p in equivalent_urls[1:]:
        assert driver._normalize_path(p) == first_path


@pytest.mark.parametrize("provider", ["mysql"])
def test_parse_connection_string_with_invalid_schemes(provider):
    from terracotta import drivers

    invalid_schemes = (
        "fakescheme://test.example.com/foo",
        "fakescheme://test.example.com:80/foo",
    )

    for invalid_scheme in invalid_schemes:
        with pytest.raises(ValueError) as exc:
            driver = drivers.get_driver(invalid_scheme, provider)
            print(type(driver))
        assert "unsupported URL scheme" in str(exc.value)


def test_get_driver_invalid():
    from terracotta import drivers

    with pytest.raises(ValueError) as exc:
        drivers.get_driver("", provider="foo")
    assert "Unknown database provider" in str(exc.value)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_creation(driver_path, provider):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")
    db.create(keys)

    assert db.key_names == keys
    assert db.get_datasets() == {}


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_creation_descriptions(driver_path, provider):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")
    key_desc = {"some": "explanatory text with unicode µóáßð©ßéó"}
    db.create(keys, key_descriptions=key_desc)

    assert db.key_names == keys
    assert db.get_keys()["some"] == key_desc["some"]


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_creation_invalid(driver_path, provider):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("invalid keyname",)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.create(keys)

    assert "must be alphanumeric" in str(exc.value)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_creation_invalid_description(driver_path, provider):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.create(keys, key_descriptions={"unknown_key": "blah"})

    assert "contains unknown keys" in str(exc.value)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_creation_reserved_names(driver_path, provider):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("page", "limit")

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.create(keys)

    assert "key names cannot be one of" in str(exc.value)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_connect_before_create(driver_path, provider):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)

    with pytest.raises(exceptions.InvalidDatabaseError) as exc:
        with db.connect():
            pass

    assert "ran driver.create()" in str(exc.value)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_broken_connection(driver_path, provider):
    """Test whether driver can recover from a broken connection"""

    class Evanescence:
        def break_me_up_inside(self, *args, **kwargs):
            raise RuntimeError("bring me to life")

        def __getattribute__(self, key):
            return object.__getattribute__(self, "break_me_up_inside")

    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")
    db.create(keys)

    with pytest.raises(RuntimeError):
        with db.connect():
            db.meta_store._connection = Evanescence()
            db.get_keys()

    assert not db.meta_store.connected

    with db.connect():
        db.get_keys()


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_repr(driver_path, provider):
    from terracotta import drivers

    db = drivers.get_driver(driver_path, provider=provider)
    assert f"meta_store={DRIVER_CLASSES[provider]}" in repr(db)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_version_match(driver_path, provider):
    from terracotta import drivers, __version__

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)

    assert __version__ == db.db_version


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_version_conflict(driver_path, provider, raster_file, monkeypatch):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider=provider)
    keys = ("some", "keynames")

    db.create(keys)
    db.insert(["some", "value"], str(raster_file))

    # works
    with db.connect():
        pass

    with monkeypatch.context() as m:
        fake_version = "0.0.0"
        m.setattr("terracotta.__version__", fake_version)
        db.meta_store.db_version_verified = False

        with pytest.raises(exceptions.InvalidDatabaseError) as exc:
            with db.connect():
                pass

        assert fake_version in str(exc.value)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_invalid_key_types(driver_path, provider):
    from terracotta import drivers, exceptions

    db = drivers.get_driver(driver_path, provider)
    keys = ("some", "keys")

    db.create(keys)

    db.get_datasets()
    db.get_datasets(["a", "b"])
    db.get_datasets({"some": "a", "keys": "b"})
    db.get_datasets(None)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.get_datasets(45)
    assert "unknown key type" in str(exc)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.delete(["a"])
    assert "wrong number of keys" in str(exc)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.delete(None)
    assert "wrong number of keys" in str(exc)

    with pytest.raises(exceptions.InvalidKeyError) as exc:
        db.get_datasets({"not-a-key": "val"})
    assert "unrecognized keys" in str(exc)


@pytest.mark.parametrize("provider", TESTABLE_DRIVERS)
def test_use_credentials_from_settings(driver_path, provider, monkeypatch):
    with monkeypatch.context() as m:
        m.setenv("TC_SQL_USER", "foo")
        m.setenv("TC_SQL_PASSWORD", "bar")

        from terracotta import drivers, update_settings

        update_settings()

        if "sqlite" not in provider:
            meta_store_class = drivers.load_driver(provider)
            assert meta_store_class._parse_path("").username == "foo"
            assert meta_store_class._parse_path("").password == "bar"

            driver_path_without_credentials = driver_path[driver_path.find("@") + 1 :]
            db = drivers.get_driver(driver_path_without_credentials, provider)
            assert db.meta_store.url.username == "foo"
            assert db.meta_store.url.password == "bar"



import pytest
from click.testing import CliRunner


@pytest.mark.parametrize("level", ["debug", "info"])
def test_logging_visible(level):
    from terracotta.scripts.cli import cli

    @cli.command("dummy")
    def dummy():
        import logging

        logger = logging.getLogger("terracotta")
        logger.info("test")

    runner = CliRunner()
    result = runner.invoke(cli, ["--loglevel", level, "dummy"])
    assert result.exit_code == 0
    assert "[+]\x1b[0m test\x1b[0m" in result.output


@pytest.mark.parametrize("level", ["warning", "error", "critical"])
def test_logging_invisible(level):
    from terracotta.scripts.cli import cli

    @cli.command("dummy")
    def dummy():
        import logging

        logger = logging.getLogger("terracotta")
        logger.info("test")

    runner = CliRunner()
    result = runner.invoke(cli, ["--loglevel", level, "dummy"])
    assert result.exit_code == 0
    assert not result.output


def test_entrypoint(monkeypatch, capsys):
    import sys
    from terracotta.scripts.cli import entrypoint

    with monkeypatch.context() as m:
        m.setattr(sys, "argv", ["terracotta"])
        with pytest.raises(SystemExit) as exc:
            entrypoint()
            assert exc.code == 0

    captured = capsys.readouterr()
    assert "Usage:" in captured.out


def test_entrypoint_exception(monkeypatch, capsys):
    import sys
    from terracotta.scripts.cli import cli, entrypoint

    dummy_error_message = "Dummy error message"

    @cli.command("dummy")
    def dummy():
        raise RuntimeError(dummy_error_message)

    with monkeypatch.context() as m:
        m.setattr(sys, "argv", ["terracotta", "dummy"])
        with pytest.raises(SystemExit) as exc:
            entrypoint()
            assert exc.code == 1

    captured = capsys.readouterr()
    assert "Uncaught exception" in captured.err
    assert dummy_error_message in captured.err



from marshmallow import ValidationError
import pytest
from terracotta.server.singleband import SinglebandOptionSchema


@pytest.mark.parametrize(
    "args, expected",
    [("[0, 1]", [0, 1]), ('["p2", "p28"]', ["p2", "p28"]), (None, None)],
)
def test_serde(args, expected):
    args = {"stretch_range": args}
    option_schema = SinglebandOptionSchema()
    loaded = option_schema.load(args)
    assert loaded["stretch_range"] == expected
    dumped = option_schema.dump(loaded)
    assert dumped["stretch_range"] == expected


def test_serde_validation():
    option_schema = SinglebandOptionSchema()
    with pytest.raises(ValidationError) as exc_info:
        args = {"stretch_range": '["t2", "p28"]'}
        option_schema.load(args)
    assert "Percentile format is `p<digits>`" in str(exc_info.value)

    with pytest.raises(ValidationError) as exc_info:
        args = {"stretch_range": '[{}, "p28"]'}
        option_schema.load(args)
    assert "Must be a string or a number" in str(exc_info.value)


def test_serde_bad_type():
    option_schema = SinglebandOptionSchema()
    with pytest.raises(ValidationError) as exc_info:
        dump_args = {"stretch_range": [0, {"bad": "type"}]}
        option_schema.dump(dump_args)
    assert "Must be a string or a number" in str(exc_info.value)

    with pytest.raises(ValidationError) as exc_info:
        load_args = {"stretch_range": '[0, {"bad": "type"}]'}
        option_schema.load(load_args)
    assert "Must be a string or a number" in str(exc_info.value)



"""xyz.py

Utilities to work with XYZ Mercator tiles.
"""

from typing import Optional, Sequence, Union, Mapping, Tuple, Any

import mercantile

from terracotta import exceptions
from terracotta.drivers.terracotta_driver import TerracottaDriver


# TODO: add accurate signature if mypy ever supports conditional return types
def get_tile_data(
    driver: TerracottaDriver,
    keys: Union[Sequence[str], Mapping[str, str]],
    tile_xyz: Optional[Tuple[int, int, int]] = None,
    *,
    tile_size: Tuple[int, int] = (256, 256),
    preserve_values: bool = False,
    asynchronous: bool = False,
) -> Any:
    """Retrieve raster image from driver for given XYZ tile and keys"""

    if tile_xyz is None:
        # read whole dataset
        return driver.get_raster_tile(
            keys,
            tile_size=tile_size,
            preserve_values=preserve_values,
            asynchronous=asynchronous,
        )

    # determine bounds for given tile
    metadata = driver.get_metadata(keys)
    wgs_bounds = metadata["bounds"]

    tile_x, tile_y, tile_z = tile_xyz

    if not tile_exists(wgs_bounds, tile_x, tile_y, tile_z):
        raise exceptions.TileOutOfBoundsError(
            f"Tile {tile_z}/{tile_x}/{tile_y} is outside image bounds"
        )

    mercator_tile = mercantile.Tile(x=tile_x, y=tile_y, z=tile_z)
    target_bounds = mercantile.xy_bounds(mercator_tile)

    return driver.get_raster_tile(
        keys,
        tile_bounds=target_bounds,
        tile_size=tile_size,
        preserve_values=preserve_values,
        asynchronous=asynchronous,
    )


def tile_exists(bounds: Sequence[float], tile_x: int, tile_y: int, tile_z: int) -> bool:
    """Check if an XYZ tile is inside the given physical bounds."""
    mintile = mercantile.tile(bounds[0], bounds[3], tile_z)
    maxtile = mercantile.tile(bounds[2], bounds[1], tile_z)

    return mintile.x <= tile_x <= maxtile.x and mintile.y <= tile_y <= maxtile.y



def test_generate_cmaps(tmpdir):
    from terracotta.cmaps.get_cmaps import SUFFIX
    from terracotta.cmaps.generate_cmaps import generate_maps

    generate_maps(str(tmpdir))
    assert (tmpdir / f"jet{SUFFIX}").check(file=True)



"""Benchmarks for all performance-critical use cases

Run separately via `pytest tests/benchmarks.py`.
"""

import pytest
from click.testing import CliRunner

ZOOM_XYZ = {"birds-eye": 8, "balanced": 12, "subpixel": 21, "preview": None}


@pytest.fixture(scope="session")
def benchmark_database(big_raster_file_nodata, big_raster_file_mask, tmpdir_factory):
    from terracotta import get_driver, update_settings

    keys = ["type", "band"]

    update_settings(RASTER_CACHE_SIZE=0)

    dbpath = tmpdir_factory.mktemp("db").join("db-readonly.sqlite")
    driver = get_driver(dbpath, provider="sqlite")
    driver.create(keys)

    mtd = driver.compute_metadata(str(big_raster_file_nodata))

    with driver.connect():
        driver.insert(["nodata", "1"], str(big_raster_file_nodata), metadata=mtd)
        driver.insert(["nodata", "2"], str(big_raster_file_nodata), metadata=mtd)
        driver.insert(["nodata", "3"], str(big_raster_file_nodata), metadata=mtd)
        driver.insert(["mask", "1"], str(big_raster_file_mask), metadata=mtd)

    return dbpath


def get_xyz(raster_file, zoom):
    import rasterio
    import rasterio.warp
    import mercantile

    with rasterio.open(str(raster_file)) as src:
        raster_bounds = rasterio.warp.transform_bounds(
            src.crs, "epsg:4326", *src.bounds
        )
    raster_center_x = (raster_bounds[0] + raster_bounds[2]) / 2
    raster_center_y = (raster_bounds[1] + raster_bounds[3]) / 2

    tile = mercantile.tile(raster_center_x, raster_center_y, zoom)
    return (tile.x, tile.y, zoom)


@pytest.mark.parametrize("resampling", ["nearest", "linear", "cubic", "average"])
@pytest.mark.parametrize("zoom", ZOOM_XYZ.keys())
def test_bench_rgb(
    benchmark, zoom, resampling, big_raster_file_nodata, benchmark_database
):
    from terracotta.server import create_app
    from terracotta import update_settings

    update_settings(
        DRIVER_PATH=str(benchmark_database),
        RESAMPLING_METHOD=resampling,
        REPROJECTION_METHOD=resampling,
    )

    zoom_level = ZOOM_XYZ[zoom]

    flask_app = create_app()
    with flask_app.test_client() as client:
        if zoom_level is not None:
            x, y, z = get_xyz(big_raster_file_nodata, zoom_level)
            rv = benchmark(client.get, f"/rgb/nodata/{z}/{x}/{y}.png?r=1&g=2&b=3")
        else:
            rv = benchmark(client.get, "/rgb/nodata/preview.png?r=1&g=2&b=3")

    assert rv.status_code == 200


def test_bench_rgb_out_of_bounds(benchmark, big_raster_file_nodata, benchmark_database):
    from terracotta.server import create_app
    from terracotta import update_settings

    update_settings(DRIVER_PATH=str(benchmark_database))

    x, y, z = 0, 0, 20

    flask_app = create_app()
    with flask_app.test_client() as client:
        rv = benchmark(client.get, f"/rgb/nodata/{z}/{x}/{y}.png?r=1&g=2&b=3")

    assert rv.status_code == 200


@pytest.mark.parametrize("resampling", ["nearest", "linear", "cubic", "average"])
@pytest.mark.parametrize("zoom", ZOOM_XYZ.keys())
def test_bench_singleband(
    benchmark, zoom, resampling, big_raster_file_nodata, benchmark_database
):
    from terracotta.server import create_app
    from terracotta import update_settings, get_driver

    update_settings(
        DRIVER_PATH=str(benchmark_database),
        RESAMPLING_METHOD=resampling,
        REPROJECTION_METHOD=resampling,
    )

    zoom_level = ZOOM_XYZ[zoom]

    flask_app = create_app()
    with flask_app.test_client() as client:
        if zoom_level is not None:
            x, y, z = get_xyz(big_raster_file_nodata, zoom_level)
            rv = benchmark(client.get, f"/singleband/nodata/1/{z}/{x}/{y}.png")
        else:
            rv = benchmark(client.get, "/singleband/nodata/1/preview.png")

    assert rv.status_code == 200
    assert not len(get_driver(str(benchmark_database)).raster_store._raster_cache)


def test_bench_singleband_out_of_bounds(benchmark, benchmark_database):
    from terracotta.server import create_app
    from terracotta import update_settings

    update_settings(DRIVER_PATH=str(benchmark_database))

    x, y, z = 0, 0, 20

    flask_app = create_app()
    with flask_app.test_client() as client:
        rv = benchmark(client.get, f"/singleband/nodata/1/{z}/{x}/{y}.png")

    assert rv.status_code == 200


@pytest.mark.parametrize("chunks", [False, True])
@pytest.mark.parametrize("raster_type", ["nodata", "masked"])
def test_bench_compute_metadata(
    benchmark, big_raster_file_nodata, big_raster_file_mask, chunks, raster_type
):
    from terracotta import raster

    if raster_type == "nodata":
        raster_file = big_raster_file_nodata
    elif raster_type == "masked":
        raster_file = big_raster_file_mask

    benchmark(raster.compute_metadata, str(raster_file), use_chunks=chunks)


@pytest.mark.parametrize("in_memory", [False, True])
def test_bench_optimize_rasters(benchmark, unoptimized_raster_file, tmpdir, in_memory):
    from terracotta.scripts import cli

    input_pattern = str(unoptimized_raster_file.dirpath("*.tif"))
    in_memory_flag = "--in-memory" if in_memory else "--no-in-memory"

    runner = CliRunner()
    result = benchmark(
        runner.invoke,
        cli.cli,
        [
            "optimize-rasters",
            input_pattern,
            "-o",
            str(tmpdir),
            in_memory_flag,
            "--overwrite",
        ],
    )

    assert result.exit_code == 0



"""server/metadata.py

Flask route to handle /metadata calls.
"""

from typing import Any, Mapping, Dict
import json

from marshmallow import Schema, fields, validate, pre_load, ValidationError
from flask import jsonify, Response, request

from terracotta.server.flask_api import METADATA_API
from terracotta.exceptions import InvalidArgumentsError


class MetadataSchema(Schema):
    class Meta:
        ordered = True

    keys = fields.Dict(
        keys=fields.String(),
        values=fields.String(),
        description="Keys identifying dataset",
        required=True,
    )
    bounds = fields.List(
        fields.Number(),
        validate=validate.Length(equal=4),
        required=True,
        description="Physical bounds of dataset in WGS84 projection",
    )
    convex_hull = fields.Dict(
        required=True, description="GeoJSON representation of the dataset's convex hull"
    )
    valid_percentage = fields.Number(
        description="Percentage of valid data in the dataset"
    )
    range = fields.List(
        fields.Number(),
        validate=validate.Length(equal=2),
        required=True,
        description="Minimum and maximum data value",
    )
    mean = fields.Number(description="Data mean", required=True)
    stdev = fields.Number(description="Data standard deviation", required=True)
    percentiles = fields.List(
        fields.Number(),
        validate=validate.Length(equal=99),
        required=True,
        description="1st, 2nd, 3rd, ..., 99th data percentile",
    )
    metadata = fields.Raw(
        description="Any additional (manually added) metadata", required=True
    )


class MetadataColumnsSchema(Schema):
    columns = fields.List(
        fields.String(),
        description="List of columns to return",
        required=False,
    )

    @pre_load
    def validate_columns(
        self, data: Mapping[str, Any], **kwargs: Any
    ) -> Dict[str, Any]:
        data = dict(data.items())
        var = "columns"
        val = data.get(var)
        if val:
            try:
                data[var] = json.loads(val)
            except json.decoder.JSONDecodeError as exc:
                raise ValidationError(
                    f"Could not decode value for {var} as JSON"
                ) from exc
        return data


class MultipleMetadataDatasetsSchema(Schema):
    keys = fields.List(
        fields.List(
            fields.String(),
            description="Keys identifying dataset",
            required=True,
        ),
        required=True,
        description="Array containing all available key combinations",
    )


@METADATA_API.route("/metadata/<path:keys>", methods=["GET"])
def get_metadata(keys: str) -> Response:
    """Get metadata for given dataset
    ---
    get:
        summary: /metadata
        description: Retrieve metadata for given dataset (identified by keys).
        parameters:
          - name: keys
            in: path
            description: Keys of dataset to retrieve metadata for (e.g. 'value1/value2')
            type: path
            required: true
          - in: query
            schema: MetadataColumnsSchema
        responses:
            200:
                description: All metadata for given dataset
                schema: MetadataSchema
            404:
                description: No dataset found for given key combination
    """
    from terracotta.handlers.metadata import metadata

    columns_schema = MetadataColumnsSchema()
    columns = columns_schema.load(request.args).get("columns")

    parsed_keys = [key for key in keys.split("/") if key]

    payload = metadata(columns, parsed_keys)
    schema = MetadataSchema(partial=columns is not None)
    return jsonify(schema.load(payload))


@METADATA_API.route("/metadata", methods=["POST"])
def get_multiple_metadata() -> Response:
    """Get metadata for multiple datasets
    ---
    post:
        summary: /metadata
        description:
            Retrieve metadata for multiple datasets, identified by the
            body payload. Desired columns can be filtered using the ?columns
            query.
        parameters:
          - in: query
            schema: MetadataColumnsSchema
          - in: body
            schema: MultipleMetadataDatasetsSchema
        responses:
            200:
                description: All metadata for given dataset
                schema: MetadataSchema
            400:
                description:
                    If the maximum number of requested datasets is exceeded
            404:
                description: No dataset found for given key combination
    """
    from terracotta.handlers.metadata import multiple_metadata

    request_body = request.json
    if not isinstance(request_body, dict):
        raise InvalidArgumentsError("Request body must be a JSON object")

    datasets_schema = MultipleMetadataDatasetsSchema()
    datasets = datasets_schema.load(request_body).get("keys")

    columns_schema = MetadataColumnsSchema()
    columns = columns_schema.load(request.args).get("columns")

    payload = multiple_metadata(columns, datasets)
    schema = MetadataSchema(many=True, partial=columns is not None)
    return jsonify(schema.load(payload))



from setuptools import setup, find_packages
from os import path
import re

here = path.abspath(path.dirname(__file__))

# get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# resolve relative links to figures
long_description = re.sub(
    r"!\[(.*?)\]\((docs/_figures/.*?)\)",
    r"![\1](https://github.com/DHI-GRAS/terracotta/raw/main/\2)",
    long_description,
)

numpy_version = ">=1.15,!=1.17.0"

setup(
    # metadata
    name="terracotta",
    description="A modern XYZ tile server written in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DHI-GRAS/terracotta",
    author="Jonas Sølvsteen",
    author_email="josl@dhigroup.com",
    keywords="xyz tileserver rasterio cloud-optimized-geotiff serverless",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Flask",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Multimedia :: Graphics :: Viewers",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    # module
    packages=find_packages(exclude=["docs", "tests"]),
    python_requires=">=3.8",
    use_scm_version={"write_to": "terracotta/_version.py"},
    # dependencies
    setup_requires=[
        "setuptools_scm",
        "setuptools_scm_git_archive",
        "numpy%s" % numpy_version,
    ],
    install_requires=[
        "apispec>=1.0",
        "apispec-webframeworks",
        "cachetools>=3.1.0",
        "click",
        "click-spinner",
        "flask",
        "flask_cors",
        "marshmallow>=3.0.0",
        "mercantile",
        "numpy%s" % numpy_version,
        "pillow",
        "pyyaml>=3.10",  # downstream dependency of apispec
        "shapely",
        "rasterio>=1.3.0",
        "shapely",
        "sqlalchemy>=1.4.1",
        "toml",
        "tqdm",
    ],
    extras_require={
        "test": [
            "pytest",
            "pytest-cov",
            "pytest-benchmark",
            "attrs>=17.4.0",
            "codecov",
            "colorlog",
            "crick",
            "matplotlib",
            "moto",
            "aws-xray-sdk",
            "pymysql>=1.0.0",
            "psycopg2",
        ],
        "docs": [
            "sphinx",
            "sphinx_autodoc_typehints",
            "sphinx-click",
            "pymysql>=1.0.0",
            "psycopg2",
        ],
        "recommended": ["colorlog", "crick", "pymysql>=1.0.0", "psycopg2"],
    },
    # CLI
    entry_points="""
        [console_scripts]
        terracotta=terracotta.scripts.cli:entrypoint
    """,
    # package data
    include_package_data=True,
    package_data={
        "terracotta": [
            "cmaps/data/*_rgba.npy",  # colormaps
            "templates/*.html",
            "static/*",  # preview app
        ]
    },
)



"""raster.py

Extract information from raster files through rasterio.
"""

from typing import Optional, Any, Dict, Tuple, Sequence, TYPE_CHECKING
import contextlib
import warnings
import logging

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from rasterio.io import DatasetReader  # noqa: F401

try:
    from crick import TDigest, SummaryStats

    has_crick = True
except ImportError:  # pragma: no cover
    has_crick = False

from terracotta import exceptions
from terracotta.profile import trace

logger = logging.getLogger(__name__)


def convex_hull_candidate_mask(mask: np.ndarray) -> np.ndarray:
    """Returns a reduced boolean mask to speed up convex hull computations.

    Exploits the fact that only the first and last elements of each row and column
    can contribute to the convex hull of a dataset.
    """
    assert mask.ndim == 2
    assert mask.dtype == np.bool_

    nx, ny = mask.shape
    out = np.zeros_like(mask)

    # these operations do not short-circuit, but seems to be the best we can do
    # NOTE: argmax returns 0 if a slice is all True or all False
    first_row = np.argmax(mask, axis=0)
    last_row = nx - 1 - np.argmax(mask[::-1, :], axis=0)
    first_col = np.argmax(mask, axis=1)
    last_col = ny - 1 - np.argmax(mask[:, ::-1], axis=1)

    all_rows = np.arange(nx)
    all_cols = np.arange(ny)

    out[first_row, all_cols] = out[last_row, all_cols] = True
    out[all_rows, first_col] = out[all_rows, last_col] = True

    # filter all-False slices
    out &= mask

    return out


def compute_image_stats_chunked(dataset: "DatasetReader") -> Optional[Dict[str, Any]]:
    """Compute statistics for the given rasterio dataset by looping over chunks."""
    from rasterio import features, warp, windows
    from shapely import geometry

    total_count = valid_data_count = 0
    tdigest = TDigest()
    sstats = SummaryStats()
    convex_hull = geometry.Polygon()

    block_windows = [w for _, w in dataset.block_windows(1)]

    for w in block_windows:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="invalid value encountered.*")
            block_data = dataset.read(1, window=w, masked=True)

        # handle NaNs for float rasters
        block_data = np.ma.masked_invalid(block_data, copy=False)

        total_count += int(block_data.size)
        valid_data = block_data.compressed()

        if valid_data.size == 0:
            continue

        valid_data_count += int(valid_data.size)

        if np.any(block_data.mask):
            hull_candidates = convex_hull_candidate_mask(~block_data.mask)
            hull_shapes = [
                geometry.shape(s)
                for s, _ in features.shapes(
                    np.ones(hull_candidates.shape, "uint8"),
                    mask=hull_candidates,
                    transform=windows.transform(w, dataset.transform),
                )
            ]
        else:
            w, s, e, n = windows.bounds(w, dataset.transform)
            hull_shapes = [geometry.Polygon([(w, s), (e, s), (e, n), (w, n)])]
        convex_hull = geometry.MultiPolygon([convex_hull, *hull_shapes]).convex_hull

        tdigest.update(valid_data)
        sstats.update(valid_data)

    if sstats.count() == 0:
        return None

    convex_hull_wgs = warp.transform_geom(
        dataset.crs, "epsg:4326", geometry.mapping(convex_hull)
    )

    return {
        "valid_percentage": valid_data_count / total_count * 100,
        "range": (sstats.min(), sstats.max()),
        "mean": sstats.mean(),
        "stdev": sstats.std(),
        "percentiles": tdigest.quantile(np.arange(0.01, 1, 0.01)),
        "convex_hull": convex_hull_wgs,
    }


def compute_image_stats(
    dataset: "DatasetReader", max_shape: Optional[Sequence[int]] = None
) -> Optional[Dict[str, Any]]:
    """Compute statistics for the given rasterio dataset by reading it into memory."""
    from rasterio import features, warp, transform
    from shapely import geometry

    out_shape = (dataset.height, dataset.width)

    if max_shape is not None:
        out_shape = (min(max_shape[0], out_shape[0]), min(max_shape[1], out_shape[1]))

    data_transform = transform.from_bounds(
        *dataset.bounds, height=out_shape[0], width=out_shape[1]
    )
    raster_data = dataset.read(1, out_shape=out_shape, masked=True)

    if dataset.nodata is not None:
        # nodata values might slip into output array if out_shape < dataset.shape
        raster_data = np.ma.masked_equal(raster_data, dataset.nodata, copy=False)

    # handle NaNs for float rasters
    raster_data = np.ma.masked_invalid(raster_data, copy=False)

    valid_data = raster_data.compressed()

    if valid_data.size == 0:
        return None

    if np.any(raster_data.mask):
        hull_candidates = convex_hull_candidate_mask(~raster_data.mask)
        hull_shapes = (
            geometry.shape(s)
            for s, _ in features.shapes(
                np.ones(hull_candidates.shape, "uint8"),
                mask=hull_candidates,
                transform=data_transform,
            )
        )
        convex_hull = geometry.MultiPolygon(hull_shapes).convex_hull
    else:
        # no masked entries -> convex hull == dataset bounds
        w, s, e, n = dataset.bounds
        convex_hull = geometry.Polygon([(w, s), (e, s), (e, n), (w, n)])

    convex_hull_wgs = warp.transform_geom(
        dataset.crs, "epsg:4326", geometry.mapping(convex_hull)
    )

    return {
        "valid_percentage": valid_data.size / raster_data.size * 100,
        "range": (float(valid_data.min()), float(valid_data.max())),
        "mean": float(valid_data.mean()),
        "stdev": float(valid_data.std()),
        "percentiles": np.percentile(valid_data, np.arange(1, 100)),
        "convex_hull": convex_hull_wgs,
    }


@trace("compute_metadata")
def compute_metadata(
    path: str,
    *,
    extra_metadata: Optional[Any] = None,
    use_chunks: Optional[bool] = None,
    max_shape: Optional[Sequence[int]] = None,
    large_raster_threshold: Optional[int] = None,
    rio_env_options: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    import rasterio
    from rasterio import warp
    from terracotta.cog import validate

    row_data: Dict[str, Any] = {}
    extra_metadata = extra_metadata or {}

    if max_shape is not None and len(max_shape) != 2:
        raise ValueError("max_shape argument must contain 2 values")

    if use_chunks and max_shape is not None:
        raise ValueError("Cannot use both use_chunks and max_shape arguments")

    if rio_env_options is None:
        rio_env_options = {}

    with rasterio.Env(**rio_env_options):
        if not validate(path):
            warnings.warn(
                f"Raster file {path} is not a valid cloud-optimized GeoTIFF. "
                "Any interaction with it will be significantly slower. Consider optimizing "
                "it through `terracotta optimize-rasters` before ingestion.",
                exceptions.PerformanceWarning,
                stacklevel=3,
            )

        with rasterio.open(path) as src:
            if src.nodata is None and not has_alpha_band(src):
                warnings.warn(
                    f"Raster file {path} does not have a valid nodata value, "
                    "and does not contain an alpha band. No data will be masked."
                )

            bounds = warp.transform_bounds(
                src.crs, "epsg:4326", *src.bounds, densify_pts=21
            )

            if (
                use_chunks is None
                and max_shape is None
                and large_raster_threshold is not None
            ):
                use_chunks = src.width * src.height > large_raster_threshold

                if use_chunks:
                    logger.debug(
                        f"Computing metadata for file {path} using more than "
                        f"{large_raster_threshold // 10**6}M pixels, iterating "
                        "over chunks"
                    )

            if use_chunks and not has_crick:
                warnings.warn(
                    "Processing a large raster file, but crick failed to import. "
                    "Reading whole file into memory instead.",
                    exceptions.PerformanceWarning,
                )
                use_chunks = False

            if use_chunks:
                raster_stats = compute_image_stats_chunked(src)
            else:
                raster_stats = compute_image_stats(src, max_shape)

    if raster_stats is None:
        raise ValueError(f"Raster file {path} does not contain any valid data")

    row_data.update(raster_stats)

    row_data["bounds"] = bounds
    row_data["metadata"] = extra_metadata
    return row_data


def get_resampling_enum(method: str) -> Any:
    from rasterio.enums import Resampling

    if method == "nearest":
        return Resampling.nearest

    if method == "linear":
        return Resampling.bilinear

    if method == "cubic":
        return Resampling.cubic

    if method == "average":
        return Resampling.average

    raise ValueError(f"unknown resampling method {method}")


def has_alpha_band(src: "DatasetReader") -> bool:
    from rasterio.enums import MaskFlags, ColorInterp

    return (
        any([MaskFlags.alpha in flags for flags in src.mask_flag_enums])
        or ColorInterp.alpha in src.colorinterp
    )


@trace("get_raster_tile")
def get_raster_tile(
    path: str,
    *,
    reprojection_method: str = "nearest",
    resampling_method: str = "nearest",
    tile_bounds: Optional[Tuple[float, float, float, float]] = None,
    tile_size: Tuple[int, int] = (256, 256),
    preserve_values: bool = False,
    target_crs: str = "epsg:3857",
    rio_env_options: Optional[Dict[str, Any]] = None,
) -> np.ma.MaskedArray:
    """Load a raster dataset from a file through rasterio.

    Heavily inspired by mapbox/rio-tiler
    """
    import rasterio
    from rasterio import transform, windows, warp
    from rasterio.vrt import WarpedVRT
    from affine import Affine

    dst_bounds: Tuple[float, float, float, float]

    if rio_env_options is None:
        rio_env_options = {}

    if preserve_values:
        reproject_enum = resampling_enum = get_resampling_enum("nearest")
    else:
        reproject_enum = get_resampling_enum(reprojection_method)
        resampling_enum = get_resampling_enum(resampling_method)

    with contextlib.ExitStack() as es:
        es.enter_context(rasterio.Env(**rio_env_options))
        try:
            with trace("open_dataset"):
                src = es.enter_context(rasterio.open(path))
        except OSError:
            raise IOError("error while reading file {}".format(path))

        # compute buonds in target CRS
        dst_bounds = warp.transform_bounds(src.crs, target_crs, *src.bounds)

        if tile_bounds is None:
            tile_bounds = dst_bounds

        # prevent loads of very sparse data
        cover_ratio = (
            (dst_bounds[2] - dst_bounds[0])
            / (tile_bounds[2] - tile_bounds[0])
            * (dst_bounds[3] - dst_bounds[1])
            / (tile_bounds[3] - tile_bounds[1])
        )

        if cover_ratio < 0.01:
            raise exceptions.TileOutOfBoundsError("dataset covers less than 1% of tile")

        # compute suggested resolution in target CRS
        dst_transform, _, _ = warp.calculate_default_transform(
            src.crs, target_crs, src.width, src.height, *src.bounds
        )
        dst_res = (abs(dst_transform.a), abs(dst_transform.e))

        # in some cases (e.g. at extreme latitudes), the default transform
        # suggests very coarse resolutions - in this case, fall back to native tile res
        tile_transform = transform.from_bounds(*tile_bounds, *tile_size)
        tile_res = (abs(tile_transform.a), abs(tile_transform.e))

        if tile_res[0] < dst_res[0] or tile_res[1] < dst_res[1]:
            dst_res = tile_res
            resampling_enum = get_resampling_enum("nearest")

        # pad tile bounds to prevent interpolation artefacts
        num_pad_pixels = 2

        # compute tile VRT shape and transform
        dst_width = max(1, round((tile_bounds[2] - tile_bounds[0]) / dst_res[0]))
        dst_height = max(1, round((tile_bounds[3] - tile_bounds[1]) / dst_res[1]))
        vrt_transform = transform.from_bounds(
            *tile_bounds, width=dst_width, height=dst_height
        ) * Affine.translation(-num_pad_pixels, -num_pad_pixels)
        vrt_height, vrt_width = (
            dst_height + 2 * num_pad_pixels,
            dst_width + 2 * num_pad_pixels,
        )

        # remove padding in output
        out_window = windows.Window(
            col_off=num_pad_pixels,
            row_off=num_pad_pixels,
            width=dst_width,
            height=dst_height,
        )

        # construct VRT
        vrt = es.enter_context(
            WarpedVRT(
                src,
                crs=target_crs,
                resampling=reproject_enum,
                transform=vrt_transform,
                width=vrt_width,
                height=vrt_height,
                add_alpha=not has_alpha_band(src),
            )
        )

        # read data
        with warnings.catch_warnings(), trace("read_from_vrt"):
            warnings.filterwarnings("ignore", message="invalid value encountered.*")
            tile_data = vrt.read(
                1, resampling=resampling_enum, window=out_window, out_shape=tile_size
            )

            # assemble alpha mask
            mask_idx = vrt.count
            mask = vrt.read(mask_idx, window=out_window, out_shape=tile_size) == 0

            if src.nodata is not None:
                mask |= tile_data == src.nodata

    return np.ma.masked_array(tile_data, mask=mask)



from typing import Any

from flask import Flask, render_template, current_app, Blueprint

client_api = Blueprint("client_api", "terracotta.client")


@client_api.route("/", methods=["GET"])
def get_map() -> Any:
    return render_template(
        "index.html", hostname=current_app.config["terracotta_hostname"]
    )


def create_app(hostname: str) -> Flask:
    client_app = Flask("terracotta.client")
    client_app.config["terracotta_hostname"] = hostname
    client_app.register_blueprint(client_api)
    return client_app



def test_datasets_handler(testdb, use_testdb):
    import terracotta
    from terracotta.handlers import datasets

    driver = terracotta.get_driver(str(testdb))
    keys = driver.key_names
    assert datasets.datasets()
    assert datasets.datasets() == [
        dict(zip(keys, pair)) for pair in driver.get_datasets().keys()
    ]

    # check key order
    assert all(tuple(ds.keys()) == keys for ds in datasets.datasets())



up_version = (0, 8)
down_version = (0, 7)

upgrade_sql = [
    "CREATE TABLE key_names (key_name TEXT PRIMARY KEY, description TEXT, idx INTEGER UNIQUE)",
    "INSERT INTO key_names (key_name, description, idx) SELECT key, description, row_number() over (order by (select NULL)) FROM keys",
    "ALTER TABLE datasets RENAME COLUMN filepath TO path",
    "UPDATE terracotta SET version='0.8.0'",
]



def test_metadata_handler(use_testdb):
    from terracotta.handlers import metadata, datasets

    ds = datasets.datasets()[0]
    md = metadata.metadata(None, ds)
    assert md
    assert md["metadata"] == ["extra_data"]

    md = metadata.metadata(["metadata", "bounds"], ds)
    assert md
    assert len(md.keys()) == 3
    assert all(k in md.keys() for k in ("metadata", "bounds", "keys"))


def test_multiple_metadata_handler(use_testdb):
    from terracotta.handlers import metadata, datasets

    ds = datasets.datasets()
    ds1 = list(ds[0].values())
    ds2 = list(ds[1].values())

    md = metadata.multiple_metadata(None, [ds1, ds2])

    assert md
    assert md[0]["metadata"] == ["extra_data"]
    assert len(md) == 2

    md = metadata.multiple_metadata(["metadata", "bounds"], [ds1, ds2])
    assert md
    assert len(md[0].keys()) == 3
    assert all(k in md[0].keys() for k in ("metadata", "bounds", "keys"))



import pytest

import sys
from textwrap import dedent

import numpy as np


OPERANDS = {
    "v1": np.ma.masked_array(np.arange(1, 6), dtype="float64"),
    "v2": np.ma.masked_array(
        2 * np.arange(1, 6), dtype="float64", mask=np.array([1, 1, 1, 0, 0])
    ),
}


VALID_EXPR = (
    # identity
    ("v1", OPERANDS["v1"]),
    # multiline
    (
        dedent(
            """
        (
            v1 +
            v1
        )
        """
        ),
        2 * OPERANDS["v1"],
    ),
    # negation
    ("-v1", -OPERANDS["v1"]),
    # abs
    ("abs(v1)", np.abs(OPERANDS["v1"])),
    # sqrt and *
    ("sqrt(v1 * v1)", OPERANDS["v1"]),
    # sqrt and **
    ("sqrt(v1 ** 2)", OPERANDS["v1"]),
    # /
    ("v1 / v1", np.ones_like(OPERANDS["v1"])),
    # %
    ("v1 % v1", np.zeros_like(OPERANDS["v1"])),
    # simple index calculation
    (
        "(v2 - v1) / (v2 + v1)",
        (OPERANDS["v2"] - OPERANDS["v1"]) / (OPERANDS["v2"] + OPERANDS["v1"]),
    ),
    # conditionals
    ("where(v2 > v1, 100, 0)", np.where(OPERANDS["v2"] > OPERANDS["v1"], 100, 0)),
    # comparisons
    ("v1 == v2", OPERANDS["v1"] == OPERANDS["v2"]),
    ("v1 != v2", OPERANDS["v1"] != OPERANDS["v2"]),
    ("v1 > v2", OPERANDS["v1"] > OPERANDS["v2"]),
    ("v1 < v2", OPERANDS["v1"] < OPERANDS["v2"]),
    ("v1 >= v2", OPERANDS["v1"] >= OPERANDS["v2"]),
    ("v1 <= v2", OPERANDS["v1"] <= OPERANDS["v2"]),
    ("(v1 < 0.5) & (v2 > 0.5)", (OPERANDS["v1"] < 0.5) & (OPERANDS["v2"] > 0.5)),
    ("(v1 < 0.5) | (v2 > 0.5)", (OPERANDS["v1"] < 0.5) | (OPERANDS["v2"] > 0.5)),
    ("~(v1 < 0.5) & (v2 > 0.5)", ~(OPERANDS["v1"] < 0.5) & (OPERANDS["v2"] > 0.5)),
    # maximum
    ("maximum(v1, v2)", np.maximum(OPERANDS["v1"], OPERANDS["v2"])),
    # minimum
    ("minimum(v1, v2)", np.minimum(OPERANDS["v1"], OPERANDS["v2"])),
    # sin / arcsin
    ("arcsin(sin(v1))", np.arcsin(np.sin(OPERANDS["v1"]))),
    # trigonometry
    ("sin(pi * v1)", np.sin(np.pi * OPERANDS["v1"])),
    # mask operations
    (
        "setmask(v1, getmask(v2))",
        np.ma.masked_array(OPERANDS["v1"], mask=OPERANDS["v2"].mask),
    ),
    ("setmask(v2, nomask)", np.ma.masked_array(OPERANDS["v2"], mask=np.ma.nomask)),
    (  # replaces mask
        "setmask(v2, ~getmask(v2))",
        np.ma.masked_array(OPERANDS["v2"], mask=~OPERANDS["v2"].mask),
    ),
    (  # adds to mask
        "masked_where(~getmask(v2), v2)",
        np.ma.masked_array(OPERANDS["v2"], mask=True),
    ),
    # long expression
    ("+".join(["v1"] * 1000), sum(OPERANDS["v1"] for _ in range(1000))),
)


INVALID_EXPR = (
    # haxx
    ('__builtins__["dir"]', "not allowed in expressions"),
    # uses list
    ("[0] * 1000000000", "not allowed in expressions"),
    # uses dict
    ("{}", "not allowed in expressions"),
    # uses string
    ('"general kenobi!"', "not allowed in expressions"),
    # if construct
    ("if True: v1", "is not a valid expression"),
    # inline comparison
    ("v1 if True else 0", "not allowed in expressions"),
    # unsupported unary operator
    ("not v1", "unary operator Not"),
    # unsupported binary operator
    ("v1 ^ v1", "binary operator BitXor"),
    # and
    ("v1 and v2", "not allowed in expressions"),
    # does not return an array
    ("0", "does not return an array"),
    # more than one expression
    ("v1; v1", "is not a valid expression"),
    # dunder method
    ("__name__", "unrecognized name '__name__'"),
    # builtins
    ("dir(v1)", "unrecognized name 'dir'"),
    # method call
    ("v1.mean()", "not allowed in expressions"),
    # attribute access
    ("v1.size", "not allowed in expressions"),
    # chained comparisons
    ("where(v1 < v2 == v2, 0, 1)", "not supported"),
    # unknown operand
    ("v100", "unrecognized name 'v100'"),
    # wrong number of arguments
    ("maximum(v1, v1, v1, v1, v1)", "got 5, expected 2"),
    # not a valid expression
    ("k = v1", "is not a valid expression"),
    # internal numpy error (mismatching types)
    ("v1 & v2", "unexpected error"),
    # code injection (serious haxx)
    (
        dedent(
            """
               (lambda fc=(
                   lambda n: [
                       c for c in
                           ().__class__.__bases__[0].__subclasses__()
                           if c.__name__ == n
                       ][0]
                   ):
                   fc("function")(
                       fc("code")(
                           0,0,0,0,"KABOOM",(),(),(),"","",0,""
                       ),{}
                   )()
               )()
               """
        ),
        "not allowed in expressions",
    ),
)


@pytest.mark.parametrize("case", VALID_EXPR)
def test_valid_expression(case):
    from terracotta.expressions import evaluate_expression

    # make sure we have enough recursion depth for long expression
    sys.setrecursionlimit(10_000)

    expr, result = case

    np.testing.assert_array_equal(evaluate_expression(expr, OPERANDS), result)


@pytest.mark.parametrize("case", INVALID_EXPR)
def test_invalid_expression(case):
    from terracotta.expressions import evaluate_expression

    expr, exc_msg = case

    with pytest.raises(ValueError) as raised_exc:
        evaluate_expression(expr, OPERANDS)

    assert exc_msg in str(raised_exc.value)


def test_invalid_compop(monkeypatch):
    from terracotta.expressions import evaluate_expression, ExpressionParser

    expr = "v0 < v1"
    exc_msg = "comparison operator"

    with monkeypatch.context() as m:
        m.setattr(ExpressionParser, "NODE_TO_COMPOP", {})

        with pytest.raises(ValueError) as raised_exc:
            evaluate_expression(expr, OPERANDS)

        assert exc_msg in str(raised_exc.value)


def test_timeout():
    from terracotta.expressions import evaluate_expression

    with pytest.raises(RuntimeError) as raised_exc:
        evaluate_expression(
            "+".join(["v1"] * 100), {"v1": np.ones((256, 256))}, timeout=0
        )

    assert "timeout" in str(raised_exc.value)


def test_mask_invalid():
    from terracotta.expressions import evaluate_expression

    res = evaluate_expression("where(v1 + v2 < 10, nan, 0)", OPERANDS)
    mask = (
        (OPERANDS["v1"] + OPERANDS["v2"] < 10)
        | OPERANDS["v1"].mask
        | OPERANDS["v2"].mask
    )

    assert isinstance(res, np.ma.MaskedArray)
    assert np.all(res == 0)
    assert np.array_equal(res.mask, mask)


def test_out_dtype():
    from terracotta.expressions import evaluate_expression

    operands = dict(v1=np.ones(10, dtype="int64"), v2=np.zeros(10, dtype="int32"))
    res = evaluate_expression("v1 + v2", operands)

    assert isinstance(res, np.ma.MaskedArray)
    assert res.dtype == np.dtype("int64")



from PIL import Image
import numpy as np

import pytest

from terracotta import exceptions


def test_rgb_handler(use_testdb, raster_file, raster_file_xyz):
    import terracotta
    from terracotta.handlers import rgb

    raw_img = rgb.rgb(["val21", "x"], ["val22", "val23", "val24"], raster_file_xyz)
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == (*terracotta.get_settings().DEFAULT_TILE_SIZE, 3)


def test_rgb_tile_size(use_testdb, raster_file, raster_file_xyz):
    from terracotta.handlers import rgb

    raw_img = rgb.rgb(
        ["val21", "x"],
        ["val22", "val23", "val24"],
        raster_file_xyz,
        tile_size=(100, 100),
    )
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == (100, 100, 3)


def test_rgb_invalid_keys(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(["val21", "x", "y", "z"], ["val22", "val23", "val24"], raster_file_xyz)


def test_rgb_invalid_args(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(
            ["val21", "x"],
            ["val22", "val23", "val24"],
            raster_file_xyz,
            stretch_ranges=[],
        )


def test_rgb_invalid_rgb_values(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(["val21", "x"], ["val22", "val23"], raster_file_xyz)


def test_rgb_out_of_bounds(use_testdb, raster_file):
    import terracotta
    from terracotta.handlers import rgb

    with pytest.raises(terracotta.exceptions.TileOutOfBoundsError) as excinfo:
        rgb.rgb(["val21", "x"], ["val22", "val23", "val24"], (10, 0, 0))
        assert "data covers less than" not in str(excinfo.value)


def test_rgb_lowzoom(use_testdb, raster_file, raster_file_xyz_lowzoom):
    import terracotta
    from terracotta.handlers import rgb

    with pytest.raises(terracotta.exceptions.TileOutOfBoundsError) as excinfo:
        rgb.rgb(["val21", "x"], ["val22", "val23", "val24"], raster_file_xyz_lowzoom)
        assert "data covers less than" in str(excinfo.value)


@pytest.mark.parametrize(
    "stretch_range",
    [
        [0, 20000],
        [10000, 20000],
        [-50000, 50000],
        [100, 100],
    ],
)
def test_rgb_stretch(stretch_range, use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import rgb

    ds_keys = ["val21", "x", "val22"]

    raw_img = rgb.rgb(
        ds_keys[:2],
        ["val22", "val23", "val24"],
        raster_file_xyz,
        stretch_ranges=[stretch_range] * 3,
    )
    img_data = np.asarray(Image.open(raw_img))[..., 0]

    # get unstretched data to compare to
    driver = terracotta.get_driver(testdb)

    with driver.connect():
        tile_data = get_tile_data(
            driver, ds_keys, tile_xyz=raster_file_xyz, tile_size=img_data.shape
        )

    # filter transparent values
    valid_mask = ~tile_data.mask
    assert np.all(img_data[~valid_mask] == 0)

    valid_img = img_data[valid_mask]
    valid_data = tile_data.compressed()

    assert np.all(valid_img[valid_data < stretch_range[0]] == 1)
    stretch_range_mask = (valid_data > stretch_range[0]) & (
        valid_data < stretch_range[1]
    )
    assert np.all(valid_img[stretch_range_mask] >= 1)
    assert np.all(valid_img[stretch_range_mask] <= 255)
    assert np.all(valid_img[valid_data > stretch_range[1]] == 255)


def test_rgb_invalid_stretch(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    stretch_range = [100, 0]
    ds_keys = ["val21", "x", "val22"]

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(
            ds_keys[:2],
            ["val22", "val23", "val24"],
            raster_file_xyz,
            stretch_ranges=[stretch_range] * 3,
        )


def test_rgb_percentile_stretch(use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import rgb

    ds_keys = ["val21", "x", "val22"]
    bands = ["val22", "val23", "val24"]
    pct_stretch_range = ["p2", "p98"]

    raw_img = rgb.rgb(
        ds_keys[:2],
        bands,
        raster_file_xyz,
        stretch_ranges=[pct_stretch_range] * 3,
    )
    img_data = np.asarray(Image.open(raw_img))[..., 0]

    # get unstretched data to compare to
    driver = terracotta.get_driver(testdb)

    with driver.connect():
        tile_data = get_tile_data(
            driver, ds_keys, tile_xyz=raster_file_xyz, tile_size=img_data.shape
        )
        band_metadata = driver.get_metadata(ds_keys)

        stretch_range = [
            band_metadata["percentiles"][1],
            band_metadata["percentiles"][97],
        ]

    # filter transparent values
    valid_mask = ~tile_data.mask
    assert np.all(img_data[~valid_mask] == 0)

    valid_img = img_data[valid_mask]
    valid_data = tile_data.compressed()

    assert np.all(valid_img[valid_data < stretch_range[0]] == 1)
    stretch_range_mask = (valid_data > stretch_range[0]) & (
        valid_data < stretch_range[1]
    )
    assert np.all(valid_img[stretch_range_mask] >= 1)
    assert np.all(valid_img[stretch_range_mask] <= 255)
    assert np.all(valid_img[valid_data > stretch_range[1]] == 255)


@pytest.mark.parametrize(
    "stretch_range_params",
    [
        ["s2", "p98", "Invalid scale value"],
        ["pp2", "p98", "Invalid percentile value"],
        ["p", "p98", "Invalid percentile value"],
        ["2", "p8", "Invalid scale value"],
        [{}, "p98", "Invalid scale value"],
        ["p-2", "p98", "Invalid percentile, out of range"],
        ["p2", "p298", "Invalid percentile, out of range"],
    ],
)
def test_rgb_invalid_percentiles(use_testdb, raster_file_xyz, stretch_range_params):
    from terracotta.handlers import rgb

    ds_keys = ["val21", "x", "val22"]
    bands = ["val22", "val23", "val24"]

    stretch_range = stretch_range_params[:2]
    with pytest.raises(exceptions.InvalidArgumentsError) as err:
        rgb.rgb(
            ds_keys[:2],
            bands,
            raster_file_xyz,
            stretch_ranges=[stretch_range] * 3,
        )
    assert stretch_range_params[2] in str(err.value)


def test_rgb_preview(use_testdb):
    import terracotta
    from terracotta.handlers import rgb

    raw_img = rgb.rgb(["val21", "x"], ["val22", "val23", "val24"])
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == (*terracotta.get_settings().DEFAULT_TILE_SIZE, 3)



"""scripts/cli.py

Entry point for CLI.
"""

from typing import Optional, Any, Mapping
import sys

import click

from terracotta.scripts.click_types import TOMLFile
from terracotta import get_settings, update_settings, logs, __version__


@click.group("terracotta", invoke_without_command=True)
@click.option(
    "-c",
    "--config",
    type=TOMLFile(),
    default=None,
    help="Update global settings from this TOML file.",
)
@click.option(
    "--loglevel",
    help="Set level for log messages",
    default=None,
    type=click.Choice(["debug", "info", "warning", "error", "critical"]),
)
@click.version_option(version=__version__)
@click.pass_context
def cli(
    ctx: click.Context,
    config: Optional[Mapping[str, Any]] = None,
    loglevel: Optional[str] = None,
) -> None:
    """The command line interface for the Terracotta tile server.

    All flags must be passed before specifying a subcommand.

    Example:

        $ terracotta -c config.toml connect localhost:5000

    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())

    # update settings from config file
    if config is not None:
        update_settings(**config)

    # setup logging
    settings = get_settings()

    if loglevel is None:
        loglevel = settings.LOGLEVEL

    logs.set_logger(loglevel, catch_warnings=True)


def entrypoint() -> None:
    try:
        cli(obj={})
    except Exception:
        import logging

        logger = logging.getLogger(__name__)
        logger.exception("Uncaught exception!", exc_info=True)
        sys.exit(1)


from terracotta.scripts.connect import connect

cli.add_command(connect)

from terracotta.scripts.ingest import ingest

cli.add_command(ingest)

from terracotta.scripts.optimize_rasters import optimize_rasters

cli.add_command(optimize_rasters)

from terracotta.scripts.serve import serve

cli.add_command(serve)

from terracotta.scripts.migrate import migrate

cli.add_command(migrate)

if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python3

import os
import re
import glob

import tqdm
import boto3
s3 = boto3.resource('s3')

import terracotta as tc

# settings
DB_NAME = 'terracotta.sqlite'
RASTER_GLOB = r'/path/to/rasters/*.tif'
RASTER_NAME_PATTERN = r'(?P<sensor>\w{2})_(?P<tile>\w{5})_(?P<date>\d{8})_(?P<band>\w+).tif'
KEYS = ('sensor', 'tile', 'date', 'band')
KEY_DESCRIPTIONS = {
    'sensor': 'Sensor short name',
    'tile': 'Sentinel-2 tile ID',
    'date': 'Sensing date',
    'band': 'Band or index name'
}
S3_BUCKET = 'tc-testdata'
S3_RASTER_FOLDER = 'rasters'
S3_PATH = f's3://{S3_BUCKET}/{S3_RASTER_FOLDER}'

driver = tc.get_driver(DB_NAME)

# create an empty database if it doesn't exist
if not os.path.isfile(DB_NAME):
    driver.create(KEYS, KEY_DESCRIPTIONS)

# sanity check
assert driver.key_names == KEYS

available_datasets = driver.get_datasets()
raster_files = list(glob.glob(RASTER_GLOB))
pbar = tqdm.tqdm(raster_files)

for raster_path in pbar:
    pbar.set_postfix(file=raster_path)

    raster_filename = os.path.basename(raster_path)

    # extract keys from filename
    match = re.match(RASTER_NAME_PATTERN, raster_filename)
    if match is None:
        raise ValueError(f'Input file {raster_filename} does not match raster pattern')

    keys = match.groups()

    # skip already processed data
    if keys in available_datasets:
        continue

    with driver.connect():
        # since the rasters will be served from S3, we need to pass the correct remote path
        driver.insert(keys, raster_path, override_path=f'{S3_PATH}/{raster_filename}')
        s3.meta.client.upload_file(raster_path, S3_BUCKET,
                                   f'{S3_RASTER_FOLDER}/{raster_filename}')

# upload database to S3
s3.meta.client.upload_file(DB_NAME, S3_BUCKET, DB_NAME)



from click.testing import CliRunner


def test_connect(test_server):
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["connect", test_server, "--no-browser"])
    assert result.exit_code == 0


def test_connect_find_socket(test_server):
    import socket
    from contextlib import closing

    from terracotta.scripts import cli

    host = "127.0.0.1"
    port = 5556

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind((host, port))

        runner = CliRunner()
        result = runner.invoke(
            cli.cli, ["connect", test_server, "--port", port, "--no-browser"]
        )
        assert result.exit_code != 0
        assert "Could not find open port to bind to" in result.output


def test_connect_invalid_port():
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["connect", "localhost:5556", "--no-browser"])
    assert result.exit_code != 0
    assert "Could not connect" in result.output


def test_connect_invalid_path(test_server):
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["connect", f"{test_server}/foo", "--no-browser"])
    assert result.exit_code != 0
    assert "Could not connect" in result.output


def test_connect_version_mismatch(test_server, monkeypatch):
    from terracotta.scripts import cli

    fake_version = "0.0.0"

    with monkeypatch.context() as m:
        m.setattr("terracotta.scripts.connect.__version__", fake_version)

        runner = CliRunner()
        result = runner.invoke(cli.cli, ["connect", test_server, "--no-browser"])
        assert result.exit_code != 0
        assert fake_version in result.output
